#define __cplusplus__
#ifndef __NEIGHBORHOOD__
#define __NEIGHBORHOOD__

#include<iostream>
#include<istream>
#include<fstream>
#include<string>
#include<vector>
#include<sstream>
#include <cmath>

using namespace std;


/*! \class  Weighted_Neighbor
 *  \brief A class representing a weighted neighbor in a neighborhood system.
 *  \author Lemine Abdallahi
 *  \date    Aug. 2005
 */

class Weighted_Neighbor{

  
public:

  ///  index of the neighbor
  uint Index; 
  /// weight of the neighbor
  double Weight ;  


  /// Default constructor for the class Weighted_Neighbor
  Weighted_Neighbor();  

  ///  Constructor for the class Weighted_Neighbor
  Weighted_Neighbor(uint i, double w); 

  ///  Destructor for the class Weighted_Neighbor.
  ~Weighted_Neighbor();

};


// Abstract neighborhood system

/*! \class  Neighborhood_System
 *  \brief A class representing an abstract neighborhood system.
 *  \author Lemine Abdallahi
 *  \date    Aug. 2005
 */


class Neighborhood_System{
  

protected :
  /// number of sites
  uint NbPts; 
  
public :
  //  Neighborhood_System();
  virtual ~Neighborhood_System(); ///<  Destructor for the class Neighborhood_System.

  ///   Get the (weighted)  neighbors for a given site.
  virtual void GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs);

  ///   Get the maximum number of neighbors for a given site.
  uint LMAX();

  ///    Display some info about the neighborhood system.
  virtual void Info();

  // Get the number of the sites
  uint Get_N();

	///// Write to a text file
  //virtual void WriteToFile(string filename);


  ///    Read the neighborhood system definition from a text file.
  virtual void ReadFromFile(string filename); 


};


// Generic Neighborhood system

/*! \class  Generic_Neighborhood_System
 *  \brief A class representing an geenric neighborhood system.
 *  \author Lemine Abdallahi
 *  \date    Aug. 2005

 *  this is a class where all information about sites and their neighborhoods is given Explicitly.
 */


class Generic_Neighborhood_System : public Neighborhood_System{



  /// the set of all neighborhood sets.  
  vector< vector <Weighted_Neighbor *> > neighbors; 

public :
  
  /// Default constructor
  Generic_Neighborhood_System();

  /// Destructor
  ~Generic_Neighborhood_System();

  /// Free all the memory used by the object.  
  void DeleteAll(); 

  void GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs); 

  void ReadFromFile(string filename); 

  /// Writes the definition of the generic neighborhood system to a text file. weights may be discarded.
  
  void WriteToFile(string filename, bool save_weights);

  /// Writes the definition of the generic neighborhood system to a text file. weights included.
  void WriteToFile(string filename); 

  
  void Info();

};

/*! \class   Image_Local_Neighborhood 
 *  \brief A class representing a neighborhood system defined locally on an image.
 *  \author Lemine Abdallahi
 *  \date    Aug. 2005

 *  

 This is a class defining a neighborhood system on an image . 

 The neighborhood is defined locally and identically around each site. 

 A site nighbors are supposed to lie within a rectangular window. More precisely , for a given point \b I=(Ix, Iy) in the image,  its neighbors , \b J=(Jx, Jy) , verify :

           \b dlmin <= Ix - Jx <= \b dlmax

and

           \b dcmin <= Iy - Jy <= \b dcmax

 A matrix of weights is also given. 

 There are (Nlines * Ncols) sites.  Sites are numbered line by line. The site \b 0 is located on the top left corner of the image.  

 */



class Image_Local_Neighborhood : public Neighborhood_System{

  /// number of lines in the image.
  uint Nlines; 
  /// number of columns in the image.
  uint Ncols;  

  /// Upper limit of the window 
  int   dlmin;
  /// Lower limit of the window
  int  dlmax; 
  /// Left limit of the window
  int  dcmin;  
  /// Right limit of the window
  int   dcmax ; 

  /// Neighbors' weights on the window.
  vector<double> weights;

  /// the set of all neighborhood sets (precomputed).  
  vector< vector <Weighted_Neighbor *> > neighbors; 

  void precompute_neighbors();
  void Compute_Neighbors(uint site, vector< Weighted_Neighbor * > & neighs);

public :

  /// Default constructor
  Image_Local_Neighborhood (); 
  /// Constructor
  Image_Local_Neighborhood (uint Nl, uint Nc);

  /// Destructor
  ~Image_Local_Neighborhood ();

  /// Free all memory used.
  void DeleteAll();

  /// Set the neighborhood specification
  void SetWeights(vector<int> window, vector<double> wghts);

  void GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs);

  /// Read from a text file
  void ReadFromFile(string filename);
  /// Write to a text file
  void WriteToFile(string filename);
  /// Display as  a generic neighborhood.
  void DumpAsGeneric();
  void Info();
};

#endif
