#define __cplusplus__
#ifndef __RF__
#define __RF__

#include<iostream>
#include<istream>
#include<fstream>
#include<string>
#include<vector>
#include<sstream>
#include <cmath>

#include <Neighborhood_System.hpp>

using namespace std;

/*! \class Random_Field
 *  \brief A class representing an Conditional_MRF.
 *  \author Juliette Blanchet
 *  \date    Dec. 2005

 Generic Random Field model. Abstract. 


*/

class Random_Field{

protected :
  /// Underlying neighborhood system.
  Neighborhood_System *NHS;
 
  /// Mean field neighborhood \f$ \tilde Z \f$
  vector<double> Z_mf;  
	/// sums of all neighbors of a given class
  vector<double> S_mf;

	/// true if the sums have been precomputed.
  bool precomputed_smf;
	/// true if Gamma has to be estimated.
	bool fix;

  /// Precompute mean-field sums of \tilde Z (for the gradient algo.)
  virtual void Precompute_Smf();

  /// tolerence of the gradient algorithm. lower limit for gradient norm.
  double gradient_tol;
  /// maximum number of iterations for the gradient algorithm.
  uint gradient_maxiter;
  /// step vector for gradient descent
  vector<double> gradient_step;



public :

  /// Destructor
  virtual ~Random_Field();
	
	/// no parameter estimation
	void Fix();

  /// Get the number of classess/colors ...
  virtual uint Get_K();
  /// Get the number of sites
  uint Get_N();

  /// Hamiltonian
  virtual double H(uint i, uint k, vector<uint>  const &  Z);
  /// Mean-field hamiltonian, \f$ \tilde H \f$ 
  virtual double Hmf(uint i, uint k);

  /// Compute conditional probabilities given father class p and the neighbors Z. mean-field appoximation
  void Compute_PGmf(uint i, vector<double> & PG);
	/// Compute conditional probabilities given father class p and the neighbors Z. mean-field appoximation
  void Compute_PGmf(vector<double> & PG);
  /// Compute class conditional probabilities at site i given upper class l and the neighbors Z. 
  void Compute_PG(uint i, vector<uint>  const &  Z, vector<double> &  PG);

  /// Simulate the Block_MRF, one iteration. Gibbs sampler
  void Simulate(vector<uint> & Z);
  /// Simulate the Block_MRF, Gibbs sampler
  void Simulate(uint nbiter, vector<uint> & Z);

  /// Set the mean-field neighborhood,  \f$ \tilde Z  \f$
  void Set_Zmf( vector<double> const &  z );
  /// Set the mean-field neighborhood at site  \f$ i \f$
  void Set_Zmf( vector<double> const &  z , uint i);
  /// Get the mean-field neighborhood,  \f$ \tilde Z \f$
  void Get_Zmf( vector<double> & z );
  /// Set  neighborhood \f$ \tilde Z \f$
  void Set_Z( vector<uint> const &  z );

	/// Compute the gradient
  virtual void Gradient(vector<double> const & tik, vector<double> & Q);  
  /// Compute the gradient 
  void Gradient(vector<uint> const & Z, vector<double> & Q);
  /// Update model parameters in the gradient descent
  virtual void Update_Gradient(vector<double> const &   Q);
  /// Initialize the gradient descent
  void Init_Gradient_Descent();
  /// Finalize the gradient descent
  void Finalize_Gradient_Descent();
  ///Set gradient descent parameters 
  void Set_Gradient(double step, double tol, uint maxiter);
  /// Update gradient step vector
  void Update_Gradient_Step(vector<double> const & Qold, vector<double> const & Q);
  /// Perform the gradient descent
  void Gradient_Descent(vector<double> const &  tik);


  /// Read the model parameters from the stream is.
  virtual void Read_Params(istream & is);
  /// Write the model parameters to the stream os.
  virtual void Write_Params(ostream & os);
	/// Estimate parameters given fuzzy classification
	void Estimate (vector<double> const &  tik);
	/// Estimate parameters given hard classification
	void Estimate (vector<uint> const &  Z);

  /// Number of free parameters
  virtual uint Degrees_Of_Freedom();

  /// Print some info on the model.
  virtual void Info();
};

#endif
