#define __cplusplus__
#ifndef __SPATIALDATA__
#define __SPATIALDATA__

#include<iostream>
#include<istream>
#include<fstream>
#include<string>
#include<vector>
#include<sstream>
#include <cmath>

#include <Data.hpp>
#include <Neighborhood_System.hpp>

using namespace std;


/// Neighborhood types.
typedef enum {
  TYPE_SPATIAL,
  TYPE_IMAGE,
  TYPE_NONSPATIAL,
  TYPE_NB }  NeighborhoodType;




/*! \class  Spatial_Data
 *  \brief A class representing a spatial dataset.
 *  \author Lemine Abdallahi
 *  \date    Aug. 2005

 This class represents the \b spatial \b dataset as understood in this project. This is the main data access class. 

All the algorithms, models defined later will access data through an object of this class.

 A spatial dataset is given by :

 - A dataset (\ref Data) \b data.
 - A neighborhood system (abstract \ref Neighborhood_System) \b nhsys.

 These two members are private. they are only accessible through some (shortcut) methods.


 */



class Spatial_Data{
  /// Dataset 
  Data *data;
  /// Neighborhood system
  Neighborhood_System *nhsys;
  
public :

  /// Default Constructor
  Spatial_Data();
  /// Constructor
  Spatial_Data(Data *d, Neighborhood_System *nhs);

  /// Destructor
  ~Spatial_Data();
  /// Free all the memory used by this object.
  void DeleteAll();

    
  // shortcut for data access
  /// Get the observation at site \b i.
  void Get(uint i, vector<double> & obs );
  /// Get X_i[d]
  double Get(uint i, uint d);

  /// Get the label of site \b i.  
  uint GetLabel(uint i);
  /// Get all the labels
  void GetLabels(vector<uint> & labs);

  /// Get the number of sites.
  uint Get_N();
  /// Get the dimension of observations.
  uint Get_D();

  // shortcut for neighborhood structure
  //  Get the neighborhood system.
  Neighborhood_System * Get_NS();
  /// Get the neighbors for a given site.
  void GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs);

  /// Read the structure of the spatial data set
  void ReadStructureFromFile(string filename);
  /// Read the dataset from a file
  void ReadDataFromFile(string filename);    
  /// Read the neighborhood from a file
  void ReadNeighborhoodFromFile(string filename);
  /// Read from a file
  void ReadFromFile(string filename);
  /// Write to a file
  void WriteToFile(string filename);


  /// Compute the (empirical) mean and covariance of the data 
  void Stats(vector<double> & mean, vector<double> & cov);
  /// K-means algorithm
  void K_means(uint K, vector<double> & centers, vector<uint> & labels);

  /// display some information about the spatial dataset.
  void Info();

};

#endif
