#include <Stat.hpp>
#include <Utils.hpp>
#include <LinAlg.hpp>
#include <Data.hpp>


/** 

 Default constructor for class Data

 Sets the number of observations to zero, and assumes working in dimension 1.

*/
Data::Data(){
  NbPts=0;
  Dim=1;
  X.resize(Dim*NbPts);
  for(uint j=0; j<(Dim*NbPts) ; j++) X[j]=0.0;
  labels.resize(NbPts);
  for(uint j=0; j<(NbPts) ; j++) labels[j]=0;
}


/** 

 Constructor for class Data

 @param nbp : Number of observations in the dataset.
 @param dim : dimension.

 Sets the number of observations to nbp, and assumes working in dimension dim.

*/

Data::Data(uint nbp, uint dim){
  NbPts=nbp;
  Dim=dim;
  X.resize(Dim*NbPts);
  for(uint j=0; j<(Dim*NbPts) ; j++) X[j]=0.0;
  labels.resize(NbPts);
  for(uint j=0; j<(NbPts) ; j++) labels[j]=0;
}

/** 

 Destructor for class Data

 Frees the memory used by the observation and labels associated.

*/


Data::~Data(){  
  X.clear();
  labels.clear();
}
  
// Getters

/**
Get the number of observations.

@return The number of observations in the dataset.

*/
uint Data::Get_N(){
  return(NbPts);
}


/**
Get the dimension.

@return The dimension of the dataset.

*/

uint Data::Get_D(){
  return(Dim);
}


/**
Get the observation at the site i.

@param  i : a site.
@param  obs : a vector containing the observation at the site i.

*/


void Data::Get(uint i, vector<double> & obs ){
  if(i<NbPts){
    uint idx=i*Dim;
    obs.clear();
    obs.resize(Dim);
    for(uint j=0; j<Dim ; j++) 
      obs[j] = X[idx+j];
  }
}


/**

 Get X_i[d]

*/

double Data::Get(uint i, uint d){
  double x=0.0;
  if((i<NbPts) && (d<Dim)){
    x=X[i*Dim+d];
  }
  else{
    cerr << "Critical error in Data::Get(uint i, uint d) . Index out of bounds." << endl;
    exit(0);
  }
  return(x);
}


/**
Get the label at the site i.

@param  i : a site.
@return the label at the site i.

*/


uint Data::GetLabel(uint i){
  uint sz=labels.size();
  uint k=0;
  if(i<sz) k=labels[i];
  return(k);
}


/**
Get the labels.

@return the labels at all the sites.

*/


void Data::GetLabels(vector<uint> & labs){
  uint sz=labels.size();
  labs.clear();
  labs.resize(sz);
  for(uint i=0;i<NbPts;i++) labs[i]=labels[i]; 
}

  
// Setters


/**
Set the observation at the site i.

@param  i : a site.
@param obs : a vector containing the new values of the observation at the site i.

*/


void Data::Set(uint i, vector<double>  const &  obs ){
  if((i<NbPts)  && (Dim==obs.size()) ){
    uint idx=i*Dim;
    for(uint j=0; j<Dim ; j++) 
      X[idx+j]=obs[j];
  } 
}


/**
Set the label at the site i.

@param  i : a site.
@param  lab : the new label at the site i.

*/

void Data::SetLabel(uint i, uint lab ){
  uint sz=labels.size();
  if(i<sz) labels[i]=lab;  
}

// IO

/**
Read data from a text file.

@param  filename : File to read the data from.

The data must be read from the file sequentially. 

The file  must contain a sequence of dim*nbPts doubles. Dimension is assumed to be 1.

*/

void Data::ReadFromFile(string filename){
   // read 

  ReadFromFile(filename, Dim);
  
}


/**
Read data from a text file in a specific dimension.

@param  filename : File to read the data from.
@param  dim : Dimension

The data must be read from the file sequentially. 

The file  must contain a sequence of dim*nbPts doubles. Dimension is set to dim.

*/


// IO
void Data::ReadFromFile(string filename, uint dim){
   // read 
  
  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "Data::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool read_error=false;

  double value;
  NbPts=0;
  Dim=dim;
  X.clear();
  while( !ifs.eof()){
    if(ifs >> value){
      NbPts++;
      X.push_back(value);
    }
    else{
      if(!ifs.eof()) read_error=true;
      break;
    }
  }

  if(!read_error ) NbPts/=Dim;
  else{
    cerr << "Data::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Possible format error. Exiting." << endl;
    exit(1);    
  }

}


// IO

/**
Write data to a text file.

@param  filename : File to be created.

The dataset is written the file sequentially. 

The file  will contain a sequence of dim*nbPts double values. Each observation is written on a separated line. This format is compatible with the format required by @ref ReadFromFile 

*/


void Data::WriteToFile(string filename){
  
  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "Data::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
    
  bool write_error=false;
  for(uint i=0; i<NbPts ; i++){
    for(uint j=0; j<Dim ; j++){
      double val=X[i*Dim+j];
      if ( ! ( ofs << val << " " ) ) write_error=true ;      
    }
    if ( ! ( ofs << endl ) ) write_error=true ;    
  }
  ofs.close();
  if(write_error){
    cerr << "Data::WriteToFile : " << endl;
    cerr << "Problem writing file : " << filename << endl;
    cerr << "Possible file format error. Exiting." << endl;
    exit(1);    
  }
}

/**
   Compute simple statistics from the data (mean and covariance).
   
   Useful for estimating a gaussain distribution.

*/

void Data::Stats(vector<double> & mean, vector<double> & cov){
  mean.clear();
  cov.clear();

  mean.resize(Dim);
  cov.resize(Dim*Dim);

  for(uint d1=0; d1<Dim ; d1++){
    for(uint i=0; i<NbPts ; i++){
      mean[d1]+=X[i*Dim+d1];
    }
    mean[d1]/=NbPts;
  }
  
  for(uint d1=0; d1<Dim ; d1++){
    for(uint d2=0; d2<Dim ; d2++){    
      for(uint i=0; i<NbPts ; i++){
	double xd1=X[i*Dim+d1];
	double xd2=X[i*Dim+d2];
	cov[d1*Dim+d2]+=(xd1-mean[d1])*(xd2-mean[d2]);
      }
      cov[d1*Dim+d2]/=NbPts;
    }
  }
}


/// K-means algorithm

void Data::K_means(uint K, vector<double> & centers, vector<uint> & labels){
  uint D=Dim;
  uint N=NbPts;
  centers.resize(K*D);
  labels.resize(N);
  
  vector<uint>  old_labels(N);
  vector<uint>  sizes(K);


  /// draw centers at random
  vector<uint> sites;
  choose(N,K,sites);
//  WriteVector<uint>(cout,sites);
  for(uint k=0; k<K; k++){
    for(uint d=0; d<D ; d++){
      centers[k*D+d]=X[sites[k]*D+d];
    }
  }
   
  uint iter=0;

  uint lab_dist=1;

  while(lab_dist > 0 ){

    for(uint i=0;i<N;i++) old_labels[i]=labels[i];

    // assign labels
    for(uint i=0;i<N;i++){
      vector<double> X_i;
      vector<double> center_k(D);
      Get(i,X_i);
      uint nearest_k=0;
      for(uint d=0; d<D ; d++) center_k[d]= centers[nearest_k*D+d];
      double dist_nearest=Matrix_dist(X_i,center_k);
      for(uint k=1; k<K; k++){
				for(uint d=0; d<D ; d++) center_k[d]= centers[k*D+d];
				double cdist=Matrix_dist(X_i,center_k);
				if(cdist<=dist_nearest){
					nearest_k=k;
					dist_nearest=cdist;
				}
      }
      labels[i]=nearest_k;
      X_i.clear();      
    }

    // recompute centers
    
    for(uint k=0;k<K;k++){
      for(uint d=0; d<D ; d++) centers[k*D+d]=0.0;
      sizes[k]=0;
    }
    
    for(uint i=0;i<N;i++){
      uint k=labels[i];
      sizes[k]+=1;
      vector<double> X_i;
      Get(i,X_i);
      for(uint d=0; d<D ; d++) centers[k*D+d]+=X_i[d];      
      X_i.clear();      
    }
    
    for(uint k=0;k<K;k++){
      if( sizes[k]!=0 ){ 
				for(uint d=0; d<D ; d++) centers[k*D+d]/=sizes[k];
			}
      else{
				cerr << "KMeans , size[k] =0" << endl;
      }
    }
        
    // conmpute labels distance.
    lab_dist=0;
    for(uint i=0;i<N;i++) if( old_labels[i]!=labels[i]) lab_dist+=1;
        
    iter++;
    
  }
}



/**
Display some info about the dataset.

This method is here for verification purposes. It displays a set of (statistical ) descriptors about the data.  

Current descriptos are: the size, dimension, column means , standard deviations, maxima , minima

*/

void Data::Info(){
  cout << "-------------------------------------------- " << endl;
  cout << "Data Information : " << endl;
  cout << "Data Points : " << NbPts << endl;
  cout << "Dimension : " << Dim << endl;

  Summary(X,Dim);
}


