#include <Stat.hpp>
#include <LinAlg.hpp>
#include <Utils.hpp>
#include <Diag_Normal.hpp>

Diag_Normal::Diag_Normal(){
  Dim=1;
  Mu.resize(Dim);
  Sigma.resize(Dim);
  for(uint d=0;d<Dim;d++){
    Mu[d]=0.0;
    Sigma[d]=1.0;
  }  
	fix=false;  
}

Diag_Normal::Diag_Normal(uint dim){
  Dim=dim;
  Mu.resize(Dim);
  Sigma.resize(Dim);
  for(uint d=0;d<Dim;d++){
    Mu[d]=0.0;
    Sigma[d]=1.0;
  }    
	fix=false;
} 

Diag_Normal::Diag_Normal(uint dim, vector<double> const &  mu, vector<double> const &  sigma){
  Dim=dim;
  Mu.resize(Dim);
  Sigma.resize(Dim);
  for(uint idx=0;idx<(Dim); idx++) Mu[idx]=mu[idx];
  for(uint idx=0;idx<(Dim); idx++) Sigma[idx]=sigma[idx];

	fix=false;
}

Diag_Normal::~Diag_Normal(){
	DeleteAll();
} 

void Diag_Normal::DeleteAll(){
  Mu.clear();
  Sigma.clear();
}

void  Diag_Normal::Simulate(vector<double> & X){

  X.clear();
  rnorm_diag(Mu,Sigma,X);
}


void  Diag_Normal::Get_Mu(vector<double> &mu){
  mu.clear();
  mu.resize(Dim);
  for (uint d=0;d<Dim;d++) mu[d]=Mu[d];
}

void  Diag_Normal::Get_Sigma(vector<double> &sigma){
  sigma.clear();
  sigma.resize(Dim);
  for (uint d=0;d<Dim;d++) sigma[d]=Sigma[d];
}

double Diag_Normal::LogDensity(vector<double> const &  X){
  double  LOG_TWO_PI=Log(2*M_PI);

  double logf=0.0;
  double log_det_Sigma =0.0;
  double sum=0.0;

  for (uint d=0; d< Dim ; d++){
    sum+= (X[d] - Mu[d] ) * (X[d] - Mu[d] ) / Sigma[d] ;
    log_det_Sigma += Log(Sigma[d]);
  }

  logf = logf - sum;
  logf = logf - ( Dim * LOG_TWO_PI ) ;  
  logf = logf - log_det_Sigma ;
  logf = logf * 0.5;


  return(logf);
}

void Diag_Normal::Estimate_Mu(Spatial_Data *spatd, vector<double> const &   freqs){
	if (fix) return;  

  // Number of sites
  uint N=spatd->Get_N();

  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  
  for(uint d=0 ; d < Dim ; d++){
    Mu[d]=0.0;
    for(uint i=0 ; i < N ; i++){
      double x_id=spatd->Get(i,d);	
      Mu[d]+=freqs[i]*x_id;
    }
    Mu[d]/=sum_freqs;
  }
}



void Diag_Normal::Estimate_Sigma(Spatial_Data *spatd, vector<double>  const &  freqs){

	if (fix) return;

  // Number of sites
  uint N=spatd->Get_N();
  
  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  
  for(uint d=0 ; d < Dim ; d++){
      
    double s=0.0;
    for(uint i=0 ; i < N ; i++){
      double x_id=spatd->Get(i,d);
      
      s+=freqs[i]*(x_id-Mu[d])*(x_id-Mu[d]);
    }
    s/=sum_freqs;
    
    Sigma[d]= s;
    
  }
}  



void Diag_Normal::Estimate(Spatial_Data *spatd, vector<double> const &   freqs){
  Estimate_Mu(spatd,freqs);
  Estimate_Sigma(spatd,freqs);
}


void Diag_Normal::Estimate(Spatial_Data *spatd){
  spatd->Stats(Mu,Sigma);
}


void Diag_Normal::Read_Params(istream & is){
  ReadVector<double>(is,Mu);
  Dim=Mu.size();
  ReadVector<double>(is,Sigma);
  if(Sigma.size()!=(Dim)){
    cerr << "Diag_Normal::Read_Params() " << endl;
    cerr << "Error reading a Diag_Normal distribution parametrs." << endl;
    exit(0);
  }

}

void  Diag_Normal::Write_Params(ostream & os){
  WriteVector<double>(os,Mu);
  WriteVector<double>(os,Sigma);
}



void Diag_Normal::Set_Mu(vector<double> const &  m){
  for(uint d=0;d<Dim;d++) Mu[d]=m[d];  
}

void Diag_Normal::Set_Sigma(vector<double> const &  sig){
  for(uint d=0;d<Dim;d++) Sigma[d]=sig[d];  
}



/// Number of free parameters
uint Diag_Normal::Degrees_Of_Freedom(){
  return(Dim + Dim);
}

void Diag_Normal::Info(){
  //cout << "in void HighDim_Normal::Info()" << endl;
  cout << "Diag_Normal:" << endl;
  cout << "Dim="<< Dim << endl;
  cout << "Mu="; WriteVector<double>(cout,Mu);
	cout << "Sigma="; WriteVector<double>(cout,Sigma);
}
