#include <Stat.hpp>
#include <LinAlg.hpp>
#include <Utils.hpp>
#include <External_Potts.hpp>

// default constructor
External_Potts::External_Potts(){
  K=1;
  NHS=new Image_Local_Neighborhood();
  Alpha.resize(K);
  for(uint i=0;i<K;i++) Alpha[i]=0.0;
  Beta=0.1;
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();

  double step_init=1.8/( N * K+ N*K*lmax) ;
  uint nb_param=K+1;
 	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);

  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}

// Constructor
External_Potts::External_Potts(Neighborhood_System *nhs, uint k){
  K=k;
  NHS=nhs;
  Alpha.resize(K);
  for(uint i=0;i<K;i++) Alpha[i]=0.0;
  Beta=0.1;
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();

  double step_init=1.8/( N * K+ N*K*lmax) ;
  uint nb_param=K+1;
	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);
  
  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}


// Constructor
External_Potts::External_Potts(Neighborhood_System *nhs, uint k, vector<double > const &  alph, double bet){
  Beta=bet;
  K=k;
  NHS=nhs;
  // singleton potentials set to zero if alph does not have the right size.
  if(alph.size()==K){
    Alpha.resize(K);
    for(uint i=0;i<K;i++) Alpha[i]=alph[i];
  }
  else{
    cerr << "External_Potts::External_Potts(Neighborhood_System *nhs, uint k, vector<double > alph, double bet) : " << endl << " bad initialization. Setting external field to zero. " << endl;
    Alpha.resize(K);
    for(uint i=0;i<K;i++) Alpha[i]=0.0;
  }
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();

  double step_init=1.8/( N * K+ N*K*lmax) ;
  uint nb_param=K+1;
	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);
  
  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}

// default destructor
External_Potts::~External_Potts(){
  DeleteAll();
}


void External_Potts::DeleteAll(){
  Alpha.clear();
}

void External_Potts::Get_Alpha(vector<double> & alpha){
  alpha.clear();
  alpha.resize(K);
  for(uint i=0;i<K;i++) alpha[i]=Alpha[i];
}

double External_Potts::Get_Beta(){
  return(Beta);
}

void External_Potts::Get_Beta(vector<double> & beta){
	beta.clear();
	beta.resize(1,Beta);
}


void External_Potts::ReadFromFile(string filename){
  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "External_Potts::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool read_error=false;
  
  if(!(ifs >> K)) read_error=true;  
  ReadVector<double>(ifs,Alpha);
  if((Alpha.size() != K) || (K<=0)) read_error=true;
  vector<double> b;
  ReadVector<double>(ifs,b);
  if(b.size()==1) Beta=b[0];
  if(b.size()!=1) read_error=true;
  b.clear();

  if(read_error){
    cerr << "External_Potts::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Possible format error. Exiting." << endl;
    exit(1);    
  }

}



void External_Potts::WriteToFile(string filename){
  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "External_Potts::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool write_error=false;
  
  if(!(ofs << K)) write_error=true;
  
  WriteVector<double>(ofs,Alpha);
  vector<double> b(1);
  b[0]= Beta;
  WriteVector<double>(ofs,b);
  b.clear();

  if(write_error){
    cerr << "External_Potts::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Possible format error. Exiting." << endl;
    exit(1);    
  }

}



double External_Potts::H(uint i , uint k,  vector<uint>  const &  Z){
  
  vector< Weighted_Neighbor * >  ineighs;
  ineighs.clear();
  NHS->GetNeighbors(i,ineighs);
  uint nv=ineighs.size();  

  double pot_k=0.0;
  for (uint v = 0; v<nv ; v++){
    
    uint    j   = ineighs[v]->Index;
    //    double w_ij = ineighs[v]->Weight;
    
    double z_jk;
    if(Z[j]==k) z_jk=1;
    else z_jk=0;
    
    pot_k+=  z_jk;
  }    
  
  ineighs.clear();
  return(Alpha[k]+Beta*pot_k);
}



double External_Potts::Hmf(uint i, uint k){
  double e=0.0;
  if(precomputed_smf){  
    e = Alpha[k];
    e += Beta * S_mf[i*K+k];
  }
  else{
    vector< Weighted_Neighbor * >  ineighs;
    NHS->GetNeighbors(i,ineighs);
    uint nv=ineighs.size();
    double sum=0.0;
    for(uint v=0;v<nv;v++){
      uint j=ineighs[v]->Index;
      sum += Z_mf[j*K+k];
    }
    e=Alpha[k] + Beta * sum;
    ineighs.clear(); 
  }

  return(e);
}






void External_Potts::Gradient(vector<double>  const & t, vector<double> & Q){

  //cerr << "External_Potts::Gradient() ." << endl;

  if(!precomputed_smf) Precompute_Smf();

  uint N=NHS->Get_N();

  Q.clear();
  Q.resize(K+1);

  vector<double> P_G;
  Compute_PGmf(P_G);
  //  WriteVector<double>(cerr,P_G);

  for(uint k=0; k<K ; k++){
    double sum_i=0;
    for(uint i=0;i<N;i++) sum_i += t[i*K+k] - P_G[i*K+k];
    Q[k]=sum_i;
  }

  double sum_i=0;
  for(uint p=0; p<K ; p++){
    for(uint i=0;i<N;i++){
      double sum_zip=S_mf[i*K+p];
      sum_i += sum_zip*(t[i*K+p] - P_G[i*K+p]);
    }
  }
  Q[K]=sum_i;       

  P_G.clear();
}

void External_Potts::Update_Gradient(vector<double>  const &  Q){

  for(uint k=0 ; k < K ; k++) Alpha[k]= Alpha[k] + gradient_step[k] * Q[k];
  Beta= Beta + gradient_step[K] * Q[K];  
  
}


/// Number of free parameters
uint External_Potts::Degrees_Of_Freedom(){
  return(K + 1);
}



void  External_Potts::Read_Params(istream & is){
  ReadVector<double>(is,Alpha);
  K=Alpha.size();
  vector<double> b;
  ReadVector<double>(is,b);
  if(b.size()==1) Beta=b[0];
  
  if( (K<=0) || (b.size()!=1) ){
    cerr << "External_Potts::Read_Params() " << endl;
    cerr << "Error reading  parametrs." << endl;
    exit(0);
  }
  b.clear();
}

void  External_Potts::Write_Params(ostream & os){
  WriteVector<double>(os,Alpha);
  vector<double> b(1);
  b[0]= Beta;
  WriteVector<double>(os,b);
  b.clear();
}

void External_Potts::Info(){
  cout << "External field Potts :" << endl;
  cout << "K = " << K << endl;
  cout << "Alpha = " ;
  WriteVector<double>(cout,Alpha);
  cout << "Beta = " << Beta << endl;
}
