#include <HMRF.hpp>
#include <Stat.hpp>
#include <Diag_Normal.hpp>
#include <Normal.hpp>
#include <Potts.hpp>
#include <LinAlg.hpp>
/** 

Default constructor

*/
HMRF::HMRF(){
  uint K=1;
  uint Dim=1;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=new Diag_Normal(Dim); 
  RF=new Potts(K);
}

/** 

Default constructor

*/
HMRF::HMRF(Neighborhood_System *nhs){
  uint K=1;
  uint Dim=1;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=new Diag_Normal(Dim); 
  RF=new Potts(nhs,K);
}


/** 

Default constructor

*/
HMRF::HMRF(uint k, uint dim , Neighborhood_System *nhs){
  uint K=1;
  uint Dim=dim;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=new Diag_Normal(Dim); 
  RF=new Potts(nhs,K);
}


/** 

Detailed constructor

*/
HMRF::HMRF( Random_Field *rf , vector<Distribution *> const &  distrib){

  RF = rf;
  uint K=RF->Get_K();
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]= distrib[k]; 
    
}

/** 

Detailed constructor

*/
HMRF::HMRF( Random_Field *rf , vector<Diag_Normal *> const &  distrib){


  RF = rf;
  uint K=RF->Get_K();
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]= distrib[k]; 
    
}

HMRF::HMRF( Random_Field *rf , vector<Normal *> const &  distrib){


  RF = rf;
  uint K=RF->Get_K();
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]= distrib[k]; 
    
}

HMRF::HMRF( Random_Field *rf , vector<HighDim_Normal *> const &  distrib){


  RF = rf;
  uint K=RF->Get_K();
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]= distrib[k]; 
    
}

HMRF::HMRF( Random_Field *rf , vector<Laplace *> const &  distrib){


  RF = rf;
  uint K=RF->Get_K();
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]= distrib[k]; 
    
}

/**

Destructor.

*/
HMRF::~HMRF(){
	Distrib.clear();
}


/**
Get the number of sites.

@return The number of sites.

*/

uint HMRF::Get_N(){
  return(RF->Get_N());
}


/**

Compute the Hamiltonian of the MRF.

@param  i : a site.
@param  k : a class.
@param  z : MRF state.

@return The value of the hamiltonian.

*/


double HMRF::H(uint i, uint k, vector<uint> const &  z){
  return(RF->H(i,k,z));
}

/**

Compute the MRF Hamiltonian mean-field approximation, given \f$ \tilde Z\f$.

@param  i : a site.
@param  k : a class.

@return The value of the approximate hamiltonian.

*/

double HMRF::Hmf(uint i, uint k){
  return(RF->Hmf(i,k));
}


/** 
  Set the gradient descent parameters. 

  @param[in] step : gradient descent step.
  @param[in] tol : lower limit for the gradient.
  @param[in] maxiter : maximum number of iterations.


  Defines the (optimal) parameters for the gradient descent to converge.

  At each iteration, the model parameters are updated with a step \b step.

  The gradient descent is stopped whenever the gradient is under \b tol or the maximum number of iterations is reached.
*/

void HMRF::Set_Gradient(double step, double tol, uint maxiter){
  RF->Set_Gradient(step,tol,maxiter);
}


/**
   Compute 
   
 \f$   \tilde \Pi(Z_i = k | \tilde z  ) = \frac { Exp[ \tilde H_{ik}] } { \sum_h Exp [ \tilde H_{ih}] }  , \forall k=1, \dots ,K  \f$

  @param [out] P_G : a vector containing all the probabilities.

*/

void HMRF::Compute_PGmf(vector<double> & P_G ){
  
  RF->Compute_PGmf(P_G);

}


/**

 Set the mean-field neighborhood, \f$ \tilde Z \f$

 @param[in] z : a vector of doubles.

*/

void HMRF::Set_Zmf( vector<double> const &  z ){
  RF->Set_Zmf(z);
}

/**

 Set the mean-field neighborhood, \f$ \tilde Z \f$ at site i

 @param[in] z : a vector of doubles.
 @param[in] i : a site.
 

*/

void HMRF::Set_Zmf( vector<double> const &  z , uint i){
  RF->Set_Zmf(z, i);
}

/**

 Get the mean-field neighborhood, \f$ \tilde Z \f$ .

 @param[out] z : a vector of doubles.

*/

void HMRF::Get_Zmf( vector<double> & z ){
  RF->Get_Zmf(z);
}

/** 

Simulation procedure.
    
*/

void HMRF::Simulate(Data *dat, uint gibbs_sampler_nbiter ){

  uint N= RF->Get_N();

  // simulate Potts model ... gibbs sampler

  vector<uint> field;
  RF->Simulate(gibbs_sampler_nbiter, field);
  //  WriteVector<uint>(cout,field);
  
  // simulate distributions ...
  for(uint i=0;i<N;i++){
    uint k=field[i];
    vector<double> X;
    Distrib[k]->Simulate(X);
    dat->Set(i,X);
    dat->SetLabel(i,k);
    X.clear();
  }

  field.clear();
}



/** 

Simulation procedure.
    
*/

void HMRF::Simulate(Data *dat, double grad_tol ){

  uint N= RF->Get_N();

  // simulate Potts model ... gibbs sampler

  vector<uint> field;

  vector<double> Grad(1);
  Grad[0]=grad_tol+2;
  uint gradit=0;
  for(uint ii=0; norm(Grad)>grad_tol ; ii++){
    RF->Simulate(1, field);
    RF->Gradient(field, Grad);
    gradit=ii;
  }
  cerr << "Gibbs sampler iterations :  " << gradit << " Gradient norm = " << norm(Grad) << endl;
 //  WriteVector<uint>(cout,field);
  
  // simulate distributions ...
  for(uint i=0;i<N;i++){
    uint k=field[i];
    vector<double> X;
    Distrib[k]->Simulate(X);
    dat->Set(i,X);
    dat->SetLabel(i,k);
    X.clear();
  }

  field.clear();
}


  
/**
   Comptue the Log-Likelihood


  @param[in] spatd : Spatial data (observations).

  @return the value of the log-likelihood.
*/

double HMRF::LogLikelihood(Spatial_Data *spatd){
  double p=0.0;
  uint K=Get_K();
  uint N=Get_N();

  vector<double> PG;
  Compute_PGmf( PG );
  
  for(uint i=0; i<N ; i++){
    double p_i=0.0;
    
    vector<double> X_i;
    spatd->Get(i,X_i);

    for(uint k=0; k< K ; k++){
      p_i += PG[i*K+k]* Exp(LogDensity(X_i,k));
    } 
    p+=Log(p_i);
    X_i.clear();
  }
  PG.clear();
  return(p);
}


/**
Compute the  BIC -- Bayesian Information Criterion.

  @param[in] spatd : Spatial data (observations).

  @return the value of the BIC.

  This is an mean-field approximation of the BIC.

*/

double HMRF::BIC(Spatial_Data *spatd){
  
  uint N=Get_N();
  double bic=2.0*LogLikelihood(spatd) - Degrees_Of_Freedom()*Log((double) N) ;
  return(bic);

}

/**
Compute the ICL --  Integrated Completed Likelihood

  @param[in] spatd : Spatial data (observations).

  @return the value of the ICL.

*/

/// 
double HMRF::ICL(Spatial_Data *spatd){

  double bic=BIC(spatd);
  uint K=Get_K();
  uint N=Get_N();

  vector<double> P_G;
  Compute_PGmf(P_G );
   
  //compute z_map
  vector<uint> z_map(N);
  for(uint i=0 ; i < N ; i++){
    vector<double> X_i;
    spatd->Get(i,X_i);
    double val_max=-1.0;
    uint k_max=0;
    for (uint k=0;k<K;k++){
			double loc_energy=Hmf(i,k);
			double val=loc_energy + LogDensity(X_i,k);
			if (val>val_max) {val_max=val;k_max=k;}
    }
    z_map[i]=k_max;
    X_i.clear();
  }

  double p_zmap = 0.0;
  for(uint i=0 ; i < N ; i++){

    double p_i=0;

    vector<double> X_i;
    spatd->Get(i,X_i);

    double sum=0.0;
    for(uint k=0 ; k < K ; k++){
      
      double val = P_G[i*K+k]*Exp(LogDensity(X_i,k));   
      sum += val;
      if(k == z_map[i]) p_i=val;
    }
    
    p_i=p_i/sum;

    p_zmap = p_zmap + Log(p_i);
    X_i.clear();
  }

  P_G.clear();
  double icl = bic + 2.0 * p_zmap; 
  return(icl);
}



/** 
    Estimate then MRF model given the fuzzy classification.

    @param[in] spatd : Spatial data (observations). Not used here.
    @param[in] tik : fuzzy classifications
    
*/
void HMRF::Estimate_MRF(vector<double>  const &  tik){
  
  RF->Estimate(tik);

}


/**

 Estimate the model parameters

 @param[in] spatd : Spatial data (observations). 
 @param[in] tik : fuzzy classifications

 Abstract . Model-dependent.

*/

void HMRF::Estimate(Spatial_Data *spatd, vector<double> const & tik){
  Estimate_Distributions(spatd , tik);
  Estimate_MRF(tik);
}

 
/**

 Compute the Completed log likelihood.

 @param[in] spatd : Spatial data (observations). 
 @param[in] tik : fuzzy classifications

 @return the value of the completed likelihhood.
*/

double HMRF::Completed_LogLikelihood(Spatial_Data *spatd, vector<double> const & tik){

  double eps=0.000000001;
  double p=0.0;

  uint K=Get_K();
  uint N=Get_N();

  vector<double> P_G;
  Compute_PGmf(P_G );
  
  for(uint i=0; i<N ; i++){
    double p_i=0.0;
    
    vector<double> X_i;
    spatd->Get(i,X_i);

    for(uint k=0; k< K ; k++){
      if(tik[i*K+k] > eps ) 
				p_i += tik[i*K+k]* ( Log(P_G[i*K+k]) + LogDensity(X_i,k) );
    } 
    p+=p_i;
    X_i.clear();
  }
  P_G.clear();
  return(p);
  

}


/**

 PLIC -- Pseudolikelihood Information Criterion

    @param[in] spatd : Spatial data (observations). 
    @param[in] Z_ICM : restoration labels.

    @return the value of the PLIC.

*/
double HMRF::PLIC(Spatial_Data *spatd, vector<uint> const &  Z_ICM){
  uint K=Get_K();
  uint N=Get_N();

  vector<double> z(N*K);
  for(uint i=0;i< N ;i++){
    for(uint k=0 ; k<K ; k++) {
      z[i*K+k]=0.0;
      if(Z_ICM[i]==k) z[i*K+k]=1.0;
    }
  } 
  Set_Zmf(z);

  double bic=BIC(spatd);
  z.clear();
  return(bic);
}

/**
   
Number of free parameters ... useful for model selection.


*/


uint HMRF::Degrees_Of_Freedom(){
  uint K=Get_K();
  uint d_K= RF->Degrees_Of_Freedom();
  for(uint k=0; k<K ;k++) d_K += Distrib[k]->Degrees_Of_Freedom();
  return(d_K);
}

void HMRF::Info(){
	cout << "HMRF:" << endl;
  RF->Info();
  uint K=RF->Get_K();
  for(uint k=0; k<K ; k++){
    Distrib[k]->Info();
  }
}

// I/O Functions
/**

 read  from a file 
 
 file format details :

*/

void HMRF::ReadFromFile(string filename){
  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "HMRF::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else {
		RF->Read_Params(ifs);
    uint K=Get_K();
    for(uint k=0; k<K ; k++){
      Distrib[k]->Read_Params(ifs);
    }

  }
  ifs.close();

}

/**

 save in a file 

*/

void HMRF::WriteToFile(string filename){

  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "HMRF::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else {
		RF->Write_Params(ofs);
    uint K=Get_K();
    for(uint k=0; k<K ; k++){
      Distrib[k]->Write_Params(ofs);
    }

  }

  ofs.close();
}
