#include <HMRF_Algo.hpp>
#include <Utils.hpp>
#include <Stat.hpp>

HMRF_Algorithm::~HMRF_Algorithm(){
  logdens.clear();
}

uint HMRF_Algorithm::Get_N(){
	return (SpatData->Get_N());
}


uint HMRF_Algorithm::Get_K(){
	return (Model->Get_K());
}

void HMRF_Algorithm::Write_fuzzy(string filename){
  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "Normal_MRF::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else {
    uint N=SpatData->Get_N();
    uint K=Model->Get_K();  
    ofs << fuzzy.size() << endl;
    for(uint i=0;i< N ;i++){
			vector<double> fuz(K);
      for(uint k=0; k < K ;k++) fuz[k]= fuzzy[i*K+k] ;
      WriteVector<double>(ofs,fuz);
			fuz.clear();
    }
  }
  ofs.close();
}


void HMRF_Algorithm::Init_Labels(vector<uint> const & labs){

  //cerr << "HMRF_Algorithm::Init_Labels()." << endl;
  uint N=SpatData->Get_N();
  uint K=Model->Get_K();
	//cout << fuzzy.size() << endl;
  if(labs.size()==N){
    fuzzy.clear();
    fuzzy.resize(N*K);
    for(uint i=0;i< N ;i++){
      uint k_0 = labs[i];
      for(uint k=0; k < K ;k++)	fuzzy[i*K+k]=0.0;
      fuzzy[i*K+k_0]=1.0;
    }
    Model->Set_Zmf(fuzzy);
    M_Step(); // M Step
}
  else{
    cerr << "HMRF_Algorithm::Init_Labels(). size error." << endl;        
  }
}


/**

Initialize the model using K-means algorithm labels.

*/
void HMRF_Algorithm::Init_KMeans(){
  uint K=Model->Get_K();
  vector<double> centers;
  vector<uint> labels;

  SpatData->K_means(K,centers,labels);
  centers.clear();

  Init_Labels(labels);
  labels.clear();

}

void HMRF_Algorithm::Init_fuzzy(vector<double> const &t){
  uint N=SpatData->Get_N();
  uint K=Model->Get_K();
 
  fuzzy.clear();
  fuzzy.resize(N*K);
  
  if(t.size()==(N*K)){
    
    for(uint i=0;i< N ;i++)
      for(uint k=0; k < K ;k++) 
				fuzzy[i*K+k]=t[i*K+k];
				Model->Set_Zmf(fuzzy);
				M_Step(); // M Step
  }
  else{
    cerr << "HMRF_Algorithm::Init_fuzzy(). size error." << endl;    
  }
}

void HMRF_Algorithm::Init_Random(){
  uint N=SpatData->Get_N();
  uint K=Model->Get_K();
  
  fuzzy.clear();
  fuzzy.resize(N*K);

  for(uint i=0;i< N ;i++){
    uint k_0 = rand() % K;
    for(uint k=0; k < K ;k++){
      fuzzy[i*K+k]=0.0;
    }
    fuzzy[i*K+k_0]=1.0;
  }
  Model->Set_Zmf(fuzzy);
  M_Step(); // M Step
}

void HMRF_Algorithm::Update_Density(){
	uint N=Get_N();
	uint K=Get_K();
	for (uint i=0;i<N;i++){
		vector<double> X_i;
		SpatData->Get(i,X_i);
		for (uint k=0;k<K;k++) logdens[i*K+k]=Model->LogDensity(X_i,k);
		X_i.clear();
	}
}


void HMRF_Algorithm::Run(uint nbiter){
	cerr << "Warning! HMRF_Algorithm::Run(uint nbiter) is virtual." << endl;
}

void HMRF_Algorithm::E_Step(){
	uint N=Get_N();
	uint K=Get_K();
	for (uint i=0;i<N;i++){
		vector<double> probs;
		Compute_fuzzy(i,probs);
		for (uint k=0;k<K;k++) fuzzy[i*K+k]=probs[k];
		probs.clear();
	}
}

void HMRF_Algorithm::M_Step(){
	Model->Estimate(SpatData,fuzzy);
	Update_Density();
}

void HMRF_Algorithm::Info(){
  uint N=Model->Get_N();
  uint K=Model->Get_K();

  vector<double> sum_tik(K);
  for(uint k=0 ; k < K ; k++){
    double sum=0.0;
    for(uint i=0; i<N ; i++) sum+=fuzzy[i*K+k];
    sum_tik[k]=sum;
  }

  //cout << "Algorithm Information : " << endl << "Fuzzy classification : ";
  //WriteVector<double>(cout,sum_tik);
  sum_tik.clear();
//   cout << "Fuzzy summary : " << endl;
//   Summary(fuzzy,K);
};


void HMRF_Algorithm::Compute_fuzzy(uint i,vector<double> &probs){
	cerr << "Warning!  HMRF_Algorithm::Compute_fuzzy(uint i,vector<double> &probs) is virtual." << endl; 
}

void HMRF_Algorithm::Compute_MAP_Labels(vector<uint> & Labels){

  uint N=Model->Get_N();
  uint K=Model->Get_K();
  
  Labels.clear();
  Labels.resize(N);
  
  for(uint i=0 ; i < N ; i++){
    double max_val=fuzzy[i*K+0];
    uint max_k=0;
    for(uint k=1 ; k < K ; k++){
      if(max_val<=fuzzy[i*K+k]){  
        max_val=fuzzy[i*K+k];
        max_k=k;
      }
    }
    Labels[i]=max_k;
  }
}


/// Completed likelihood criterion
void HMRF_Algorithm::Compute_CLCrit(){

  double Current_CL=Model->Completed_LogLikelihood(SpatData, fuzzy);  

  if(Crit_Values[ITER_CRIT]>0){
    double CL_diff=fabs(Old_CL-Current_CL);  
    Crit_Values[CL_CRIT]=CL_diff;
    Old_CL=Current_CL;
  }
  else{
    Old_CL=Current_CL;
  }
}

/// Fuzzy classification criterion 
void  HMRF_Algorithm::Compute_FuzzyCrit(){

  if(Crit_Values[ITER_CRIT]>0){    
    double fuzz_diff=fabs(old_fuzzy[0]-fuzzy[0]);
    for(uint idx=1 ; idx < fuzzy.size() ; idx++){
      double diff = fabs(old_fuzzy[idx]-fuzzy[idx]);
      if(fuzz_diff <=  diff )  fuzz_diff = diff;
    }
    Crit_Values[FUZZY_CRIT]=fuzz_diff;
    for(uint idx=0 ; idx < fuzzy.size() ; idx++) old_fuzzy[idx] = fuzzy[idx];
  }
  else{
    old_fuzzy.resize(fuzzy.size());
    for(uint idx=0 ; idx < fuzzy.size() ; idx++) old_fuzzy[idx] = fuzzy[idx];    
  }

}

/// Hard classification criterion
void  HMRF_Algorithm::Compute_HardCrit(){

  vector<uint> map_labs;
  Compute_MAP_Labels(map_labs);  
  if(Crit_Values[ITER_CRIT]>0){        
    double hard_diff=0.0;
    for(uint idx=0 ; idx < map_labs.size() ; idx++){
      if(old_map_labs[idx] != map_labs[idx]) hard_diff += 1.0 ;
    }
    hard_diff /= map_labs.size();
    Crit_Values[HARD_CRIT]=hard_diff;
    for(uint idx=0 ; idx < map_labs.size() ; idx++) 
      old_map_labs[idx] = map_labs[idx];
  }
  else{
    old_map_labs.resize( map_labs.size() );
    for(uint idx=0 ; idx < map_labs.size() ; idx++) 
      old_map_labs[idx] = map_labs[idx];    
  }
  map_labs.clear();

}
