#include <Stat.hpp>
#include <LinAlg.hpp>
#include <Utils.hpp>
#include <HighDim_Normal.hpp>
#include <lapack.hpp>

/*! \class HighDim_Normal
 *  \brief A class representing a highdimensional normal distribution.
 *  \author Juliette Blanchet
 *  \date   Oct. 2005

*/

HighDim_Normal::HighDim_Normal(uint dim){
  //cout << "in HighDim_Normal::HighDim_Normal(uint dim)" << endl;
  Dim=dim;
  Dim_red=1;
  A=0.0;
  B=0.0;
  Mu.resize(Dim);
  Q.resize(Dim*Dim_red);
  for(uint d1=0;d1<Dim;d1++){
    Mu[d1]=0.0;
    for(uint d2=0;d2<Dim_red;d2++){
      if (d1==d2) Q[d1*Dim_red + d2 ]=1.0;
      else Q[ d1*Dim_red + d2]=0.0;	
    }
  }
	fix=false; 
  //cout << "out HighDim_Normal::HighDim_Normal(uint dim)" << endl;   
}

HighDim_Normal::HighDim_Normal(){
  Dim=1;
  Dim_red=1;
  A=0.0;
  B=0.0;
  Mu.resize(Dim);
  Q.resize(Dim*Dim_red);
  for(uint d1=0;d1<Dim;d1++){
    Mu[d1]=0.0;
    for(uint d2=0;d2<Dim_red;d2++){
      if (d1==d2) Q[d1*Dim_red + d2 ]=1.0;
      else Q[ d1*Dim_red + d2]=0.0;	
    }
  }    
	fix=false;
}

HighDim_Normal::HighDim_Normal(uint dim, uint dim_red, double a, double b){
  Dim=dim;
  Dim_red=dim_red;
  A=a;
  B=b;
  Mu.resize(Dim);
  Q.resize(Dim*Dim_red);
  for(uint d1=0;d1<Dim;d1++){
    Mu[d1]=0.0;
    for(uint d2=0;d2<Dim_red;d2++){
      if(d1==d2) Q[d1*Dim_red + d2 ]=1.0;
      else Q[ d1*Dim_red + d2]=0.0;	
    }
  }      
	fix=false;
}

HighDim_Normal::HighDim_Normal(uint dim, uint dim_red, double a, double b, vector<double> const &  mu, vector<double> const &  q){
  Dim=dim;
  Dim_red=dim_red;
  A=a;
  B=b;
  Mu.resize(Dim);
  Q.resize(Dim*Dim_red);
  for(uint idx=0;idx<(Dim); idx++) Mu[idx]=mu[idx];
  for(uint idx=0;idx<(Dim*Dim_red); idx++) Q[idx]=q[idx];
	fix=false;
}

HighDim_Normal::~HighDim_Normal(){
  DeleteAll();
}

void HighDim_Normal::DeleteAll(){
  //cout << "in void HighDim_Normal::DeleteAll()" << endl;
  Mu.clear();
  Q.clear();
  //cout << "out void HighDim_Normal::DeleteAll()" << endl;
}


void HighDim_Normal::Info(){
  //cout << "in void HighDim_Normal::Info()" << endl;
  cout << "HighDim_Normal:" << endl;
  cout << "Dim="<< Dim << endl;
  cout << "Dim_red="<< Dim_red << endl;
  cout << "A="<< A << endl;
  cout << "B="<< B << endl;
  cout << "Q="; WriteVector<double>(cout,Q);
  //cout << "out void HighDim_Normal::Info()" << endl;
}

uint HighDim_Normal::Get_Dred(){
  return(Dim_red);
}

void HighDim_Normal::Get_Mu(vector<double> &mu){
  mu.clear();
  mu.resize(Dim);
  for(uint d=0;d<Dim;d++) mu[d]=Mu[d];
}

double HighDim_Normal::Get_A(){
  return A;
}

double HighDim_Normal::Get_B(){
  return B;
}

void HighDim_Normal::Get_Q(vector<double> &q){
  q.clear();
  q.resize(Dim*Dim_red);
  for(uint d=0;d<Dim*Dim_red;d++) q[d]=Q[d];
}

void HighDim_Normal::Get_Qc(vector<double> &qc){
  qc.clear();
  qc.resize(Dim*Dim);
  for(uint d1=0;d1<Dim;d1++){
    for (uint d2=0;d2<Dim;d2++){
      if (d2<Dim_red) qc[d1*Dim+d2]=Q[d1*Dim_red+d2];
      else qc[d1*Dim+d2]=0.0;
    }
  }
}


void HighDim_Normal::Set_Mu(vector<double> const &  m){
  for(uint d=0;d<Dim;d++) Mu[d]=m[d];  
}

void HighDim_Normal::Set_A(double a){
  A=a;  
}

void HighDim_Normal::Set_B(double b){
  B=b;  
}

void HighDim_Normal::Set_Q(vector<double> const &  q){
  for(uint d=0;d<Dim*Dim_red;d++) Q[d]=q[d];  
}

void HighDim_Normal::Compute_Delta(vector<double> Delta){
  Delta.clear();
  Delta.resize(Dim*Dim);
  for (uint d1=0;d1<Dim;d1++){
    for (uint d2=0;d2<Dim;d2++){
      if (d1==d2){
	if (d1<Dim_red) Delta[d1*Dim+d2]=A;
	else  Delta[d1*Dim+d2]=B;
      }
      else
	Delta[d1*Dim+d2]=0.0;
    }
  }
}


void HighDim_Normal::Compute_Sigma(vector<double> Sigma){
  Sigma.clear();
  Sigma.resize(Dim*Dim);
  vector<double> tQ;
  Matrix_transpose(Q,tQ);
  vector<double> Delta;
  Compute_Delta(Delta); 
  vector<double> QDelta;
  Matrix_mult(Q,Delta,QDelta);
  Matrix_mult(QDelta,tQ,Sigma);
  tQ.clear();
  Delta.clear();
  QDelta.clear();
}

void  HighDim_Normal::Simulate(vector<double> & X){

  X.clear();
  vector<double> Sigma;
  Compute_Sigma(Sigma);
  rnorm(Mu,Sigma,X);
  Sigma.clear();
}



void HighDim_Normal::Proj(vector<double> const &X, vector<double> &pX){
  //cerr << "HighDim_Normal::Proj()"<<endl;
  pX.clear();
  pX.resize(Dim);
  //vector<double> QtQ(Dim*Dim);
  for (uint d1=0;d1<Dim;d1++) {
		double temp=0.0;
    for (uint d2=0;d2<Dim;d2++){
			double sum=0.0;				
      for (uint k=0;k<Dim_red;k++) sum+=Q[d1*Dim_red+k]*Q[d2*Dim_red+k];
			//QtQ[d1*Dim+d2]=sum;
			//QtQ[d2*Dim+d1]=sum;
			temp+=sum*(X[d2]-Mu[d2]);
		}
		pX[d1]=temp+Mu[d1];
	}
	
//	for (uint d1=0;d1<Dim;d1++){
//		double sum=0.0;
//		for (uint d2=0;d2<Dim;d2++) sum+=QtQ[d1*Dim+d2]*(X[d2]-Mu[d2]);
//		pX[d1]=sum+Mu[d1];
//  }
//	QtQ.clear();

}


double HighDim_Normal::LogDensity(vector<double> const &  X){
  double  LOG_TWO_PI=Log(2*M_PI);
  
  if(X.size() != Dim) {
    cerr << "Warning. wrong dimension in HighDim_Normal::Density(vector<double> X)." << endl;
  }
  
  double logf=0.0;
  vector<double> pX;
  Proj(X,pX);
  double n1=0.0; //norm(Mu-pX);
  double n2=0.0; //norm(X-pX);
  for (uint k=0;k<Dim;k++){
    n1+=(Mu[k]-pX[k])*(Mu[k]-pX[k]);
    n2+=(X[k]-pX[k])*(X[k]-pX[k]);
  }

  logf=-0.5*(1/A*n1+1/B*n2+Dim_red*Log(A)+(Dim-Dim_red)*Log(B)+Dim*LOG_TWO_PI);
  

  pX.clear();

  return(logf);
}

void HighDim_Normal::Compute_Empirical_Sigma(Spatial_Data *spatd, vector<double>  const &  freqs, vector<double> &Sigma){

	//cerr << "void HighDim_Normal::Compute_Empirical_Sigma" << endl;

  Sigma.clear();
  Sigma.resize(Dim*Dim);
    
  //WriteVector<double>(cout,Mu);

  // Number of sites
  uint N=spatd->Get_N();
  
  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  
  for(uint d1=0 ; d1 < Dim ; d1++){
    for(uint d2=d1 ; d2 < Dim ; d2++){
      
      double s=0.0;
      for(uint i=0 ; i < N ; i++){
				double x_id1=spatd->Get(i,d1);
				double x_id2=spatd->Get(i,d2);
	
				s+=freqs[i]*(x_id1-Mu[d1])*(x_id2-Mu[d2]);
      }
      s/=sum_freqs;
      
      Sigma[d1*Dim + d2]= s;
      Sigma[d2*Dim + d1]= s; 
      
    }
  }
  //WriteVector<double>(cout,Sigma);
  
}

void HighDim_Normal::Estimate_Mu(Spatial_Data *spatd, vector<double>  const &  freqs){
  if (fix) return;
  // Number of sites
  uint N=spatd->Get_N();

  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  //WriteVector<double>(cout,freqs);
  //cout <<  "sum_freqs=" << sum_freqs << endl;
  
  for(uint d=0 ; d < Dim ; d++){
    Mu[d]=0.0;
    for(uint i=0 ; i < N ; i++){
      double x_id=spatd->Get(i,d);	
      Mu[d]+=freqs[i]*x_id;
    }
    Mu[d]/=sum_freqs;
  }   
}


void HighDim_Normal::Estimate_dabQ(Spatial_Data *spatd, vector<double>  const &  freqs){
  if (fix) return;
  // empirical sigma from data and frequencies
  vector<double> Sigma;
  Compute_Empirical_Sigma(spatd, freqs, Sigma);

  // sigma copy
  vector<double> Sigma_copy;
  Vec_copy(Sigma, Sigma_copy);

  // eigenvalues
  int lwork_eig=(32+3)*Dim;
  vector<double> work_eig(lwork_eig);
  vector<double> eigenvals(Dim);
  int info;
  int D=(int) Dim;
  dsyev_("No vector","Upper",&D,&Sigma_copy[0],&D,&eigenvals[0],&work_eig[0],&lwork_eig,&info);	
  
  // reduced dim
	double val=-1.0;
	for (uint d=0;d<Dim-1;d++){
		double sc_d=eigenvals[d+1]-eigenvals[d];
		if (sc_d>val) val=sc_d;
	}
		
  double tol=0.2; //LA ON A LE SEUIL QUI GOUVERNE LA REDUC DE DIMENSION!
  for(uint d=0;d<Dim;d++) {
		double sc_d=eigenvals[d+1]-eigenvals[d];
		if(sc_d>tol*val){
			Dim_red=Dim-d-1; 
			//Dim_red=Dim-d; 
			break;
		}
	}      
  // corresponding eigenvectors (no il to iu, ascending, inclusive)
  vector<double> eigenvecs(Dim*Dim);
  Vec_copy(Sigma,Sigma_copy);
	vector<double> eigenvals_copy;
	Vec_copy(eigenvals,eigenvals_copy);
  int il=Dim-Dim_red+1,iu=Dim;
  double dminusone;
  double dunused;
  int iunused;
  vector<int> iwork(5*Dim);
  int N=spatd->Get_N();
  vector<int> ifail(N);
           
  dsyevx_("Vectors also","Indexed values","Upper",
	  &D,&Sigma_copy[0],&D,&dunused,&dunused,&il,&iu,&dminusone,
	  &iunused,&eigenvals_copy[0],&eigenvecs[0],&D,&work_eig[0],&lwork_eig,&iwork[0],
	  &ifail[0],&info);
 
  // A
  double accu=0;
  for(uint d=Dim-Dim_red;d<Dim;d++) accu+=eigenvals[d];
  A=accu/Dim_red;
  
  // B
  double Tr=Trace(Sigma);
  B=(Tr-accu)/(Dim-Dim_red);

	cerr << "Dred=" << Dim_red << " " << "A=" << A << " " << "B=" << B << endl;
  
  // Q
  vector<double> tQ(Dim_red*Dim);
  for (uint d=0;d<Dim_red*Dim;d++) tQ[d]=eigenvecs[(Dim-Dim_red)*Dim+d];
	
  Q.clear();
  Q.resize(Dim*Dim_red);
  for (uint d1=0;d1<Dim;d1++){
    for (uint d2=0;d2<Dim_red;d2++){
      //Q[d1*Dim_red+d2]=eigenvecs[d2*Dim_red+d1];
      Q[d1*Dim_red+d2]=eigenvecs[d2*Dim+d1];
    }
  } 
  tQ.clear();
  //WriteVector<double>(cout,Q);

  Sigma_copy.clear();
  work_eig.clear();
  eigenvals.clear();
	eigenvals_copy.clear();
  eigenvecs.clear();
  iwork.clear();
  ifail.clear(); 
}

void HighDim_Normal::Estimate(Spatial_Data *spatd, vector<double>  const &  freqs){
  Estimate_Mu(spatd,freqs);
  Estimate_dabQ(spatd,freqs);
}



/// Number of free parameters
uint HighDim_Normal::Degrees_Of_Freedom(){
  return(Dim + (Dim*(Dim+1))/2);
}


void HighDim_Normal::Read_Params(istream & is){
  ReadVector<double>(is,Mu);
  Dim=Mu.size();
  is >> Dim_red;  
  is >> A;
  is >> B;
  ReadVector<double>(is,Q);
  if(Q.size()!=(Dim*Dim_red)){
    cerr << "HighDim_Normal::Read_Params() " << endl;
    cerr << "Error reading a HighDim_Normal distribution parameters." << endl;
    exit(0);
  }

}

void  HighDim_Normal::Write_Params(ostream & os){
  WriteVector<double>(os,Mu);
  os << Dim_red << endl;
  os << A << endl;
  os << B << endl;
  WriteVector<double>(os,Q);  
}

