#include <IID_Mixture.hpp>
#include <Stat.hpp>
#include <Diag_Normal.hpp>
#include <Normal.hpp>
#include <Utils.hpp>
/** 

Default constructor

*/
IID_Mixture::IID_Mixture(){
  uint K=1;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  uint Dim=1;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=new Diag_Normal(Dim);  
}

/** 

Detailed constructor

*/
IID_Mixture::IID_Mixture(uint k, uint dim ){
  uint K=k;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  uint Dim=dim;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=new Diag_Normal(Dim);

}

/** 

Detailed constructor

*/
IID_Mixture::IID_Mixture(uint k, vector<Distribution *> const &  distrib){
  uint K=k;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=distrib[k];
}

/** 

Detailed constructor

*/
IID_Mixture::IID_Mixture(uint k, vector<Diag_Normal *> const &  distrib){
  uint K=k;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=distrib[k];
}

IID_Mixture::IID_Mixture(uint k, vector<Normal *> const &  distrib){
  uint K=k;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=distrib[k];
}

IID_Mixture::IID_Mixture(uint k, vector<HighDim_Normal *> const &  distrib){
  uint K=k;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=distrib[k];
}

IID_Mixture::IID_Mixture(uint k, vector<Laplace *> const &  distrib){
  uint K=k;
  Pi.resize(K);
  for(uint k=0;k<K;k++) Pi[k]=1.0/K;
  Distrib.resize(K);
  for(uint k=0;k<K;k++) Distrib[k]=distrib[k];
}

/**

Destructor.

*/

IID_Mixture::~IID_Mixture(){
	Pi.clear();
}

/**
   Get the mixture probabilities.

   @param [out] pi : the mixture probabilities.

*/


void IID_Mixture::Get_Pi( vector<double> & pi ){
  uint K=Get_K();
	pi.clear();
  pi.resize(K);
  for(uint k=0;k<K;k++) pi[k]=Pi[k];
}

/**
   Set the mixture probabilities.

   @param [in] pi : the mixture probabilities.

*/


void IID_Mixture::Set_Pi( vector<double> const &   pi ){
  uint K=Get_K();
	if(pi.size()==K) for(uint k=0;k<K;k++) Pi[k]=pi[k];
}

  
/**
   Comptue the Log-Likelihood


  @param[in] spatd : Spatial data (observations).


  @return the value of the log-likelihood.


*/


double IID_Mixture::LogLikelihood(Spatial_Data *spatd){
  double p=0.0;
  uint N=spatd->Get_N();
  
  for(uint i=0; i<N ; i++){
    double p_i=0.0;
    
    vector<double> X_i;
    spatd->Get(i,X_i);
		uint K=Get_K();
    for(uint k=0; k< K ; k++){
      p_i += Pi[k]*Exp(LogDensity(X_i,k));
    } 
    p+=Log(p_i);
    X_i.clear();
  }
  return(p);
}

/**

 Compute the Completed likelihood.

 @param[in] spatd : Spatial data (observations). 
 @param[in] tik : fuzzy classifications

 @return the value of the completed likelihhood.


*/

double  IID_Mixture::Completed_LogLikelihood(Spatial_Data *spatd, vector<double> const &   tik){

  double eps=0.000000001;

  double p=0.0;
  uint N=spatd->Get_N();

  
  for(uint i=0; i<N ; i++){
    double p_i=0.0;
    
    vector<double> X_i;
    spatd->Get(i,X_i);
		uint K=Get_K();
    for(uint k=0; k< K ; k++){
      if(tik[i*K+k] > eps )
				p_i += tik[i*K+k]* ( Log(Pi[k]) + LogDensity(X_i,k) );
    } 
    p+=p_i;
    X_i.clear();
  }
  return(p);
  
}

/**
Compute the  BIC -- Bayesian Information Criterion.

  @param[in] spatd : Spatial data (observations).

  @return the value of the BIC.

*/


double IID_Mixture::BIC(Spatial_Data *spatd){
  
  uint N=spatd->Get_N();
  double bic=2.0*LogLikelihood(spatd) - Degrees_Of_Freedom()*Log((double) N) ;
  return(bic);

}

/**
Compute the ICL --  Integrated Completed Likelihood

  @param[in] spatd : Spatial data (observations).

  @return the value of the ICL.

Abstract. Model-dependent.


*/

double IID_Mixture::ICL(Spatial_Data *spatd){

  double bic=BIC(spatd);
  uint N=spatd->Get_N();
  uint K=Get_K();
  //compute z_map
  vector<uint> z_map(N);      
  for(uint i=0 ; i < N ; i++){
		vector<double> X_i;
    spatd->Get(i,X_i);
    double val_max=-1.0;
    uint k_max=0;
		for(uint k=0 ; k < K ; k++){
      double fki=Exp(LogDensity(X_i , k));
      double val=Pi[k]*fki;
      if (val>val_max) {k_max=k; val_max=val;}
    }
    z_map[i]=k_max;
  }

  double p_zmap = 0.0;
  for(uint i=0 ; i < N ; i++){

    double p_i=0;

    vector<double> X_i;
    spatd->Get(i,X_i);

    double sum=0.0;

    for(uint k=0 ; k < K ; k++){
      double val= Pi[k]*Exp(LogDensity(X_i,k));
      sum += val;
      if(k == z_map[i]) p_i=val;
    }
    p_i=p_i/sum;

    p_zmap = p_zmap + Log( p_i );
    X_i.clear();
  }

  double icl = bic + 2.0 * p_zmap; 
  return(icl);
}


/**

 Estimate mixture probabilities

  @param[in] tik : fuzzy classifications

  Estimate \b Pi.
*/

void IID_Mixture::Estimate_Pi(vector<double> const &   tik){
	uint K=Get_K();
  // Number of sites
  uint N=tik.size()/K;

  for(uint k=0 ; k < K ; k++){
    double sum=0.0;
    for(uint i=0 ; i < N ; i++) sum+=tik[i*K+k];
    Pi[k]=sum/N;
  }
}

void IID_Mixture::Estimate(Spatial_Data *spatd, vector<double> const & tik){
  Estimate_Pi(tik);  
	Estimate_Distributions(spatd , tik);
}

/** 

Simulation procedure.
    
*/

void IID_Mixture::Simulate(uint N, Data *dat){
  
  for(uint i=0;i<N;i++){
    uint k=Sample(Pi);
    vector<double> X;
    Distrib[k]->Simulate(X);
    dat->Set(i,X);
    dat->SetLabel(i,k);
    X.clear();
  }

}

/**
   
Number of free parameters ... useful for model selection.

*/


uint IID_Mixture::Degrees_Of_Freedom(){
  uint K=Get_K();
	uint d_K= (K-1);
  for(uint k=0; k<K ;k++) d_K += Distrib[k]->Degrees_Of_Freedom();
  return(d_K);
}

/**

 read  from a file 
 
 file format details :

*/

void IID_Mixture::ReadFromFile(string filename){
    
  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "IID_Mixture::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else {
    
    ReadVector<double>(ifs,Pi);
		uint K=Get_K();
    for(uint k=0; k<K ; k++){
      Distrib[k]->Read_Params(ifs);
    }
  }
  ifs.close();
}

/**

 save in a file 

*/

void IID_Mixture::WriteToFile(string filename){
  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "IID_Mixture::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else {
    WriteVector<double>(ofs,Pi);
		uint K=Get_K();
    for(uint k=0; k<K ; k++){
      Distrib[k]->Write_Params(ofs);
    }
  }

  ofs.close();
}



void IID_Mixture::Info(){
  cout << "IID Mixture " << endl;
  cout << "Pi =  " << endl;
  WriteVector<double>(cout,Pi);
	uint K=Get_K();
  for(uint k=0; k<K ; k++){
    Distrib[k]->Info();
  }
}




