#include <Stat.hpp>
#include <Utils.hpp>
#include <Laplace.hpp>


Laplace::Laplace(){
  Dim=1;
  Mu.resize(Dim);
  Lambda.resize(Dim);
  for(uint d=0 ; d<Dim ; d++){
    Mu[d]=0.0;
    Lambda[d]=1.0;
  } 
	fix=false;   
}

Laplace::Laplace(uint dim){
  Dim=dim;
  Mu.resize(Dim);
  Lambda.resize(Dim);
  for(uint d=0 ; d<Dim ; d++){
    Mu[d]=0.0;
    Lambda[d]=1.0;
  }    
	fix=false;
}

Laplace::Laplace(uint dim, vector<double> const &  mu, vector<double> const &  lambda){
  Dim=dim;
  Mu.resize(Dim);
  Lambda.resize(Dim);
  for(uint idx=0;idx< Dim; idx++) Mu[idx]=mu[idx];
  for(uint idx=0;idx< Dim ; idx++) Lambda[idx]=lambda[idx];
	fix=false;
}

Laplace::~Laplace(){
	DeleteAll();
}

void Laplace::DeleteAll(){
  Mu.clear();
  Lambda.clear();
}


/**

Simulate 

F^(-1) (p) = mu - lambda * sign(p-0.5) *  Log( 1 - fabs( 2 * p - 1) )

*/

void  Laplace::Simulate(vector<double> & X){

  X.clear();
  X.resize(Dim);

  for(uint d=0 ; d<Dim; d++){
    double U=runif();
    double sign=1;
    if(U<0.5) sign=-1;
    if(U==0.5) sign=0.0;
    X[d] = Mu[d] - Lambda[d] * sign * Log(1-fabs(2*U-1));    
  }
  
}

double Laplace::LogDensity(vector<double> const &  X){
  
  double logf=0;
  for(uint d=0; d<Dim; d++){
    logf += -fabs(X[d]-Mu[d])/Lambda[d] - Log(2*Lambda[d]);
  }

  return(logf);
}


void Laplace::Estimate_Mu(Spatial_Data *spatd, vector<double>  const &  freqs){
  if (fix) return;
  // Number of sites
  uint N=spatd->Get_N();

  for(uint d=0 ; d<Dim; d++){
    vector<double> X;
    X.resize(N);
    for(uint i=0 ; i < N ; i++) X[i]=spatd->Get(i,d);

    vector<double> W;
    W.resize(N);
    for(uint i=0 ; i < N ; i++) W[i]=freqs[i];
    
    Mu[d]=weighted_median(X,W);

    X.clear();
    W.clear();
  }

}



void Laplace::Estimate_Lambda(Spatial_Data *spatd, vector<double> const &   freqs){
  if (fix) return;
  // Number of sites
  uint N=spatd->Get_N();
  
  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  
  for(uint d=0 ; d < Dim ; d++){      
    double s=0.0;
    for(uint i=0 ; i < N ; i++){
      double x_id=spatd->Get(i,d);
      
      s+=freqs[i]*fabs(x_id-Mu[d]);
    }
    s/=sum_freqs;
    
    Lambda[d]= s;    
  }
  
}


void Laplace::Estimate(Spatial_Data *spatd, vector<double>  const &  freqs){
  Estimate_Mu(spatd,freqs);
  Estimate_Lambda(spatd,freqs);
}


void Laplace::Init_at_random(Spatial_Data *spatd){

  // Number of sites
  uint N=spatd->Get_N();  
  uint idx=rand() % N;
  
  for(uint d=0;d<Dim;d++){
    double x=spatd->Get(idx,d);
    Mu[d]=x;
    Lambda[d]=1.0+runif();
  }    

}


void Laplace::Read_Params(istream & is){
  ReadVector<double>(is,Mu);
  Dim=Mu.size();
  ReadVector<double>(is,Lambda);
  if(Lambda.size()!=Dim){
    cerr << "Laplace::Read_Params() " << endl;
    cerr << "Error reading a Laplace distribution parametrs." << endl;
    exit(0);
  }
}

void  Laplace::Write_Params(ostream & os){
  WriteVector<double>(os,Mu);
  WriteVector<double>(os,Lambda);
}



/// Number of free parameters
uint Laplace::Degrees_Of_Freedom(){
  return(Dim + Dim);
}

void Laplace::Info(){
  //cout << "in void HighDim_Normal::Info()" << endl;
  cout << "Laplace:" << endl;
  cout << "Dim="<< Dim << endl;
  cout << "Mu="; WriteVector<double> (cout,Mu);
	cout << "Lambda="; WriteVector<double> (cout,Lambda);
}
