#include <LinAlg.hpp>

double Trace(vector<double> const &  M){
  double trace=0.0;
  uint N2=M.size();
  uint N=(uint)sqrt((double)N2);
  if( (N*N)!=N2 ){
    cerr << "Trace on a non square matrix." << endl;
    exit(0);
  }
  else{   
    for (uint i=0;i<N;i++){trace+=M[i*N+i];}    
  }
  return trace;
}

/**

 Cholesky decomposition , M = L * L^T
 M has to be symmetric positive definite matrix.
 L is a lower triangular matrix. 
*/

void Cholesky(vector<double> const &  M, vector<double> & L){
  uint N2=M.size();
  uint N=(uint)sqrt((double)N2);
  if( (N*N)!=N2 ){
    cerr << "Cholesky on a non square matrix." << endl;
    exit(0);
  }
  else{
    L.clear();
    L.resize(N2);
    for(uint i=0; i<N2;i++) L[i] = M[i] ;
    
    for(uint i=0; i<N ;i++){
      for(uint j=i; j<N ;j++){
	double sum=L[i*N +j];
	for (uint k=0;k<i;k++)  sum -= L[i*N +k]*L[j*N +k];
	if (i == j){
	  if (sum <= 0.0){
	    cerr << "Cholesky on a non positive matrix." << endl;
	    exit(0); 
	  }
	  else L[i*N+i]=sqrt(sum);
	}
	else L[j*N +i]=sum/L[i*N+i];
      }
    }
    for(uint i=0; i<N ;i++)
      for(uint j=0; j<i ;j++) L[j*N +i]=0.0;
    
  }  
}

/** 

Determinant

L has to be a lower triangular matrix. 

In this case, The determinant is the product of diagonal elements.

*/
double Determinant(vector<double> const &  L){
  double prod=1.0;
  uint N2=L.size();
  uint N=(uint)sqrt((double)N2);
  if( (N*N)!=N2 ){
    cerr << "DEterminant of a non square matrix." << endl;
    exit(0);
  }
  else{
    for(uint i=0; i<N;i++) prod *= L [ i*N + i];
  }
  return(prod);
}



/**

 Matrix invert , L1 = L^(-1) 

 L has to be a positive lower triangular matrix. 
 

*/

void Matrix_invert(vector<double> const &  L, vector<double> & L1){
  uint N2=L.size();
  uint N=(uint)sqrt((double)N2);
  if( (N*N)!=N2 ){
    cerr << "Inverse of a non square matrix." << endl;
    exit(0);
  }
  else{
    L1.clear();
    L1.resize(N2);
    for(uint k=0; k<N2;k++) L1[k]=L[k];
 
    for(uint i=0; i<N;i++){
      L1 [i*N + i ] = 1.0 /  L [ i*N + i ];
      for(uint j=i+1; j<N;j++){
	double sum=0.0;
	for(uint k=i; k<j;k++) sum-=L1[j*N+k]*L1[k*N+i];
	L1[j*N+i]=sum/L[ j*N + j];
      }
    }
  }
}

/**

 Square matrix transpose B = A^T  

*/

void Matrix_transpose(vector<double> const &  A, vector<double> &  B ){
  uint NA2=A.size();
  uint NA=(uint)sqrt((double)NA2);
  if( (NA*NA)!=NA2 ){
    cerr << "A has to be a square matrix." << endl;
    exit(0);
  }
  else{
    B.clear();
    B.resize(NA2);
    for(uint i=0; i<NA2;i++) B[i] = 0.0 ;
    for(uint i=0; i<NA ;i++){
      for(uint j=0; j<NA ;j++){
	B[i*NA +j] = A[j*NA+i];
      }
    }
  }
}



/**

 Square matrix multiplication C = A * B  
*/

void Matrix_mult(vector<double> const &  A, vector<double> const &   B, vector<double> & C){

  uint NA2=A.size();
  uint NA=(uint)sqrt((double)NA2);
  if( (NA*NA)!=NA2 ){
    cerr << "A has to be a square matrix." << endl;
    exit(0);
  }

  uint NB2=A.size();
  uint NB=(uint)sqrt((double)NB2);
  if( (NB*NB)!=NB2 ){
    cerr << "B has to be a square matrix." << endl;
    exit(0);
  }
  
  if(NA != NB ){
    cerr << "The two matrices have to be the same size." << endl;
    exit(0);
  }
  else {
    C.clear();
    uint N=NA;
    C.resize(NA2);
    for(uint i=0; i<NA2;i++) C[i]=0.0;
    
    for(uint i=0; i<N ;i++){
      for(uint j=0; j<N ;j++){
	double sum=0.0;
	for(uint k=0; k<N ;k++) sum+=A[i*N + k]* B [k*N + j];
	C[i*N + j] = sum;
      }
    }
  }
  
}




/**

 square matrix vector multiplication. Y = A * X

*/

void Matrix_Vec_mult( vector<double> const &  A, vector<double> const & X, vector<double> & Y){
  uint NA2=A.size();
  uint NA=(uint)sqrt((double)NA2);
  if( (NA*NA)!=NA2 ){
    cerr << "Matrix-vector multiplication. A has to be a square matrix." << endl;
    exit(0);
  }
  
  uint NX=X.size();

  if(NX != NA){
    cerr << "Matrix-vector multiplication. Dimensions are not equal." << endl;
    exit(0);
  }
  else{
    Y.clear();
    Y.resize(NA);
    
    for(uint i=0; i<NA ;i++){
      Y[i]=0.0;
      for(uint j=0; j<NA ;j++) Y[i]+=A[i*NA+j]*X[j];
    }
  }
}

/**

 vector addition. v3 = v1 + v2

*/

void Vec_add( vector<double> const &  v1, vector<double> const & v2, vector<double> & v3){
  uint N1=v1.size();
  uint N2=v2.size();
  if( N1!=N2 ){
    cerr << "vector addition. v1 and v2 must have the same dimension." << endl;
    exit(0);
  }
  else{
    v3.clear();
    v3.resize(N1);
    
    for (uint d=0;d<N1;d++){
      v3[d]=v1[d]+v2[d];
    }
  }
}

/**

 vector subtraction. v3 = v1 - v2

*/

void Vec_sub( vector<double> const &  v1, vector<double> const & v2, vector<double> & v3){
  uint N1=v1.size();
  uint N2=v2.size();
  if( N1!=N2 ){
    cerr << "vector addition. v1 and v2 must have the same dimension." << endl;
    exit(0);
  }
  else{
    v3.clear();
    v3.resize(N1);
    
    for (uint d=0;d<N1;d++){
      v3[d]=v1[d]-v2[d];
    }
  }
  
}

/**

 vector copy. v2 = v1

*/

void Vec_copy(vector<double> const &  v1, vector<double> & v2){  
  uint N1=v1.size();
  v2.clear();
  v2.resize(N1);    
  for (uint d=0;d<N1;d++){
    v2[d]=v1[d];
  }
}

/** 

Print a matrix

*/

void Matrix_print(vector<double> const &  A ){

  uint NA2=A.size();
  uint NA=(uint)sqrt((double)NA2);
  if( (NA*NA)!=NA2 ){
    cerr << "A has to be a square matrix." << endl;
    exit(0);
  }
  else{
    for(uint i=0; i<NA ;i++){
      for(uint j=0; j<NA ;j++){
	cout << " " << A[i*NA + j] ;
      }
      cout << endl;
    }
  }
}


/**

 Create a random square matrix (N*N)

*/
void Matrix_random(uint N, vector<double> & A ){
  A.clear();
  A.resize(N*N);
  for(uint i=0; i<(N*N) ;i++){
    double unif=drand48();
    A[i]=unif;      
  }
  
}


/**

 Create a zero square matrix (N*N)

*/
void Matrix_setzero(uint N, vector<double> & A ){
  A.clear();
  A.resize(N*N);

  for(uint i=0; i<(N*N) ;i++) A[i]=0.0;
}

/**

 Create an identity matrix (N*N)

*/
void Matrix_setid(uint N, vector<double> & A ){
  A.clear();
  A.resize(N*N);

  for(uint i=0; i<N ;i++)
    for(uint j=0; j<N ;j++){
      if(i==j)   A[i*N+j]=1.0;
      else  A[i*N+j]=0.0;
    }
}



/**

 Matrix L2 distance 

*/
double Matrix_dist(vector<double> const &  A, vector<double> const &   B ){

  uint NA2=A.size();
  uint NB2=A.size();
  double d=0.0;
  
  if(NA2 != NB2 ){
    cerr << "The two matrices have to be the same size." << endl;
    exit(0);
  }
  else {
    for(uint i=0;i<NA2;i++) d+=(A[i]-B[i])*(A[i]-B[i]);
  }
  d=sqrt(d);
  return(d);
}


double norm(vector<double> const &  A){
  uint NA=A.size();
  double d=0.0;
  for(uint i=0 ; i<NA ; i++) d+=A[i]*A[i];
  d=sqrt(d);
  
  return(d);
}

double norm2(vector<double> const &  A){
  uint NA=A.size();
  double d=0.0;
  for(uint i=0 ; i<NA ; i++) d+=A[i]*A[i];
  
  return(d);
}

uint norm2(vector<uint> const &  A){
  uint NA=A.size();
  uint d=0;
  for(uint i=0 ; i<NA ; i++) d+=A[i]*A[i];
  
  return(d);
}
