#include <Neighborhood_System.hpp>
#include <Utils.hpp>

// Weighted_Neighbor::


/** 

  Default constructor for the class Weighted_Neighbor. 

  Creates a new object. Sets the index to zero, the weight to zero.

*/

Weighted_Neighbor::Weighted_Neighbor(){
  Index=0;
  Weight=0.0;
}


/** 

 Constructor for the class Weighted_Neighbor. 

 @param i : index.
 @param w : weight.

 Creates a new object. Sets the index to \b i, the weight to \b w.

*/


Weighted_Neighbor::Weighted_Neighbor(uint i, double w){
  Index=i;
  Weight=w;
}



/** 
 Destructor for the class Weighted_Neighbor. 

 Deletes the object. nothing else.

*/

Weighted_Neighbor::~Weighted_Neighbor(){

}

/** 
    
  Destructor for the abstract class Neighborhood_System. 

  Abstract. To be defined for the derived classes.

*/



Neighborhood_System::~Neighborhood_System(){
  
}

/** 
  Get the (weighted) neighbors for a given site. 

  @param[in] site : a site.
  @param[out]  neighs : the set of neighbors of the site \b site

  Fills the vector \b neighs with the necessary neighborhood data for the site \b site.

  Abstract. To be defined for the derived classes.

*/


void Neighborhood_System::GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs){

}



/** 
    Get the number of the sites
    
*/

uint Neighborhood_System::Get_N(){
  return(NbPts);
}


/**
   Display some info about the neighborhood system. 

   Abstract. to be defined for the derived classes.
*/


void Neighborhood_System::Info(){
  cout << "Abstract neighborhood system ..." << endl;
}


/** 
    Get the maximum number of neighbors for a given site.
    
*/

uint Neighborhood_System::LMAX(){
  uint maxNH=0;
  uint sz=Get_N();
  for(uint i=0; i< sz ; i++){
    vector< Weighted_Neighbor * >  neighs;
    GetNeighbors(i, neighs);
    uint nb= neighs.size() ;
    //    cout << "Neighborhood_System::LMAX() " << nb << endl;
    if(nb >= maxNH){
      maxNH=nb;
    }
    neighs.clear();
  }
  
  return(maxNH);
}


//void Neighborhood_System::WriteToFile(string filename){
//	cerr << "Warning. Neighborhood_System::WriteToFile(string filename) is virtual." << endl;
//}

/**
   Read the neighborhood system definition from a text file. 

   @param  filename : file from which the definition is read.

    Abstract. To be defined for the derived classes.
*/



void Neighborhood_System::ReadFromFile(string filename){
  cerr << "Warning. Neighborhood_System::ReadFromFile(string filename) is virtual." << endl;
}


/**
   Default constructor. 
   
   Sets the number of sites to zero.

*/

Generic_Neighborhood_System::Generic_Neighborhood_System(){
  NbPts=0;
}


/**
   Default destructor. 
   
   Calls \ref DeleteAll .

*/


Generic_Neighborhood_System::~Generic_Neighborhood_System(){
  DeleteAll();
}

/**

   Frees all memory allocated and used by the neighborhood information.

*/

void Generic_Neighborhood_System::DeleteAll(){
  uint sz=neighbors.size();
  for(uint i=0; i< sz ; i++) neighbors[i].clear();
  neighbors.clear();  
}




void Generic_Neighborhood_System::GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs){

  uint sz=neighbors[site].size();
  neighs.clear();
  neighs.resize(sz);
  for(uint i=0; i< sz ; i++) neighs[i]=neighbors[site][i];

}

/**
   Reads the object definition from a text file.

   \param [in] filename : name of the input file. 

   Details of the file format may be described as follows:
   
   The first line contains a number ( \b 0 or \b 1) depending on wether weights are present (\b 1)  or not (\b 0).

   The following lines give \b  for \b each  site \b i the following information

   \code
   i k j1, ....jk
   \endcode

   where \b k is the number of neighbors , \b j1 ... \b jk are the neighbors of \b i.

   and if weights are present (first line == \b 1 ) , the weights are appended to this line, this gives :

   \code
   i k j1 ... jk wi1 ... wik
   \endcode
   

   a typical example file is : 

   \code
   1  
   1 2 2 3 0.5 0.4
   2 1 1 0.7
   3 2 1 2 0.6 0.5
  \endcode

*/


void Generic_Neighborhood_System::ReadFromFile(string filename){
  // read all neighborhood system
  uint weighted=0;

  DeleteAll();

  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "Generic_Neighborhood_System::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool read_error=false;
  bool format_error=false;

  if ( ! (ifs >> weighted ) ) read_error=true;
  if ( (weighted !=0 ) && (weighted != 1)) format_error=true;

  bool first_line_read=false;
  uint min_site=0;
  while( !ifs.eof() ){
    uint     site=0 ;
    uint     site_nb_neighbors=0 ;
    
    vector<Weighted_Neighbor *> current_neighbors;
    
    current_neighbors.clear();
    //    if( ifs.eof() ) break;
    if( ! (ifs >> site ) )  read_error=true;

    if(!first_line_read){
      min_site=site;
      first_line_read=true;
    }
    if( ! ( ifs >>  site_nb_neighbors  ) )  read_error=true;

    else if( site_nb_neighbors >= 0 ) {
      vector<uint> nei_indexes;
      nei_indexes.clear();
      nei_indexes.resize(site_nb_neighbors);
      current_neighbors.resize(site_nb_neighbors);
      
      for(uint i=0; i< site_nb_neighbors; i++ ) {
				uint site_nei=0;
				if( ! ( ifs >> site_nei  ) )  read_error=true;
				nei_indexes[i]=site_nei;
      }
      if ( weighted == 1){
				double weight=0.0;
				for(uint i=0; i< site_nb_neighbors; i++ ) {
					if( ! ( ifs >> weight  ) )  read_error=true;
					Weighted_Neighbor *wn=new Weighted_Neighbor( nei_indexes[i]-min_site , weight);
					current_neighbors[i]=wn;
				}
      }
      else{
				double weight=1.0;      
				for(uint i=0; i< site_nb_neighbors; i++ ) {
					Weighted_Neighbor *wn=new Weighted_Neighbor( nei_indexes[i]-min_site , weight);
					current_neighbors[i]=wn;
				}
      }
      neighbors.push_back(current_neighbors);
      //WriteVector<uint>(cout,nei_indexes);
    }
    //    if(ifs.eof()) break;
		
    NbPts=neighbors.size();
		//cout << "NbPts=" << NbPts << " " << "nbnei=" << site_nb_neighbors << endl;
		
    if(!ifs.eof() && (read_error || format_error) ){
      cerr << "Generic_Neighborhood_System::ReadFromFile : " << endl;
      cerr << "Problem with file : " << filename << endl;
      cerr << "Possible file format error. Exiting." << endl;
      exit(1);
    }

  }  


}

/**
   Write the neighborhood definition to a text file. 

   Forces the weights presence in the output.
   
   \param [in] filename : name of the output file. 

   The format is described in \ref ReadFromFile .


*/

void Generic_Neighborhood_System::WriteToFile(string filename){
  this->WriteToFile(filename,true);
}

/**
   Write the neighborhood definition to a text file. 

   
   \param [in] filename : name of the output file. 
   \param [in] save_weights : boolean , wether to output weights or not.

   The format is described in \ref ReadFromFile .

*/


  void Generic_Neighborhood_System::WriteToFile(string filename, bool save_weights){
  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "Generic_Neighborhood_System::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool write_error=false;

  uint sz=neighbors.size();
  if(save_weights) if ( ! ( ofs << 1 << endl ) ) write_error=true ;
  else if ( ! ( ofs << 0 << endl  ) ) write_error=true ;
  for(uint i=0; i< sz ; i++){
    vector< Weighted_Neighbor * >  neighs;
    neighs.clear();
    GetNeighbors(i, neighs);
    uint sz1=neighs.size();
    if ( ! ( ofs << i << " " << sz1  ) ) write_error=true ;
    for(uint j=0; j< sz1 ; j++)  
      if ( ! ( ofs << " " << neighs[j]->Index  ) ) write_error=true ;
    if(save_weights) 
      for(uint j=0; j< sz1 ; j++)  
	 if ( ! ( ofs << " " << neighs[j]->Weight  ) ) write_error=true ;
     if ( ! ( ofs << endl  ) ) write_error=true ;
  }
  ofs.close();

  if(write_error){

    cerr << " --- Error : Generic_Neighborhood_System::WriteToFile : " << endl;
    cerr << " --- Problem with file : " << filename << endl;
    cerr << " --- Error writing file. Exiting." << endl;
    exit(1);
  }


}

/**
   Display some info about the neighborhood system.

   for verification purposes.

   Displays :
    - the number of sites
    - The site with the maximum number of neighbors.
    - The site with the minimum number of neighbors.
    - The mean number of neighbors per site.


*/


void Generic_Neighborhood_System::Info(){  
  uint sz=neighbors.size();
  cout << "-------------------------------------------- " << endl;
  cout << "Neighborhood structure information : " << endl;
  cout << "Type : Generic neighborhood. " << endl;
  cout << "Sites : " << sz << endl;
  uint _min_idx=0;
  uint _max_idx=0;
  uint _min=neighbors[0].size() ;
  uint _max=neighbors[0].size() ;
  double _mean=0.0;
  for(uint i=0; i< sz ; i++){
    vector< Weighted_Neighbor * >  neighs;
    neighs.clear();
    GetNeighbors(i, neighs);
    uint nb= neighs.size() ;
    if(nb >= _max){
      _max_idx=i;
      _max=nb;
    }
    if(nb <= _min){
      _min_idx=i;
      _min=nb;
    }
    _mean+=nb;
  }
  _mean/=sz;
  
  cout << "Max neighbors/site " << _max_idx << " " << _max << " or " << LMAX()  << endl;
  cout << "Min neighbors/site " << _min_idx << " " << _min << endl;

  cout << "Mean neighbors/site " << _mean << endl;
    
}



// Image_Local_Neighborhood 

/** 
    Default constructor. 

    Empty image, empty neighborhood. 

*/

Image_Local_Neighborhood::Image_Local_Neighborhood (){
  Nlines=0;
  Ncols=0;
  dlmin=0; 
  dlmax=0; 
  dcmin=0; 
  dcmax=0 ;
}

/** 
    Constructor. 

    \param [in] Nl : Number of columns in the image
    \param [in] Nc : Number of lines in the image

    Defines a default neighborhood system on an \b Nl * \b Nc image. 

   This is a first order local neighborhood. Each site has \b 4 neighbors (\b N, \b E, \b S, \b W). weights are all equal to \b 1.0

*/


Image_Local_Neighborhood::Image_Local_Neighborhood (uint Nl, uint Nc){

  Nlines=Nl;
  Ncols=Nc;
  NbPts=Nl*Nc;
  dlmin=-1; 
  dlmax=1; 
  dcmin=-1; 
  dcmax=1 ;
  //  uint window_size= (dlmax-dlmin+1)*(dcmax-dcmin+1) ;
  weights.push_back(0);  weights.push_back(1);
  weights.push_back(0);  weights.push_back(1);
  weights.push_back(0);  weights.push_back(1);
  weights.push_back(0);  weights.push_back(1);
  weights.push_back(0); 
  this->precompute_neighbors();
}

/**

 Set the neighborhood specification. 

\param [in] window : vector containg window limits.
\param [in] wghts : vector containing th weights on the window.

\b window is used to set (\b dlmin, \b dlmax, \b dcmin, \b dcmax) values.
Sets the weiughts according to the values of wghts.
*/

void Image_Local_Neighborhood::SetWeights(vector<int> window, vector<double> wghts){
  if(window.size()==4){  
    dlmin=window[0]; 
    dlmax=window[1]; 
    dcmin=window[2]; 
    dcmax=window[3];
    
    weights.clear();
    uint window_size= (dlmax-dlmin+1)*(dcmax-dcmin+1) ;

    if(wghts.size()==window_size ){
      for(uint i=0; i< wghts.size();i++) weights.push_back(wghts[i]);
    }
    else{
      
      cerr << "Invalid Image Local Neighborhood weights matrix specification : " ;
      for(uint i=0; i< wghts.size();i++) cerr << " " << wghts[i];
      cerr << endl;
    }
  }
  else{
    cerr << "Invalid Image Local Neighborhood window specification : " ;
    for(uint i=0; i< window.size();i++) cerr << " " << window[i];
    cerr << endl;
  }
}

/**
   Deletes the object

   Calls \ref DeleteAll .
*/

Image_Local_Neighborhood::~Image_Local_Neighborhood (){
  DeleteAll();
}


/**

  Frees all the memory used by the object.

*/

void Image_Local_Neighborhood::DeleteAll(){
  weights.clear();
  uint sz=neighbors.size();
  for(uint i=0; i< sz ; i++) neighbors[i].clear();
  neighbors.clear();  
}



void Image_Local_Neighborhood::Compute_Neighbors(uint site, vector< Weighted_Neighbor * > & neighs){
  uint line=site / Ncols ;
  uint col=site % Ncols;

  uint iv=0;
  for ( int dl = dlmin ; dl <= dlmax ; dl++ ) {
    for ( int dc = dcmin ; dc <= dcmax ; dc++ ){
      int l = line + dl ;
      int c=col+dc;
      if ( (l>=0) && (l<(int)Nlines) && (c>=0) && (c<(int)Ncols) ){
	int idx=l * Ncols + c;
	double weight=weights[iv];
	if(weight!=0.0){
	  Weighted_Neighbor *wn=new Weighted_Neighbor( idx , weight);
	  neighs.push_back(wn);
	}
      }
      iv++;
    }
  }
}

/**
   
   Precompute and store all neighborhood information.

*/
void Image_Local_Neighborhood::precompute_neighbors(){
  uint nbpts=Nlines*Ncols;
  neighbors.clear();
  neighbors.resize(nbpts);
  
  for(uint i=0; i< nbpts ;i++){
    vector< Weighted_Neighbor * >  ineighs;
    this->Compute_Neighbors(i,ineighs);
    neighbors[i]=ineighs;
    ineighs.clear();
  }  
}



/**

The set of neighbors for a given site is computed locally based on the window information and the weights associated.



*/

void Image_Local_Neighborhood::GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs){

  uint sz=neighbors[site].size();
  neighs.clear();
  neighs.resize(sz);
  for(uint i=0; i< sz ; i++) neighs[i]=neighbors[site][i];

}



/**
   Reads the object definition from a text file.

   \param [in] filename : name of the input file. 


   Details of the file format may be described as follows:
   
   the first four numbers are the windoiw specification  (\b dlmin, \b dlmax, \b dcmin, \b dcmax).

   the reamining sequence of numbers gives the weights on the window.

   

   a typical  ( correctly formatted ) example file is : 

   \code
   -1 1
   -1 1
     1    0    1
     0    0  0.3
   0.7  0.6  0.5
  \endcode

*/



void Image_Local_Neighborhood::ReadFromFile(string filename){

  DeleteAll();

  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "Image_Local_Neighborhood::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool read_failure=false;

  vector<int> window;
  vector<double> wghts;
  //  dlmin, &dlmax, &dcmin, &dcmax
  int _dlmin=0;
  if ( ! (ifs >> _dlmin ) )   read_failure=true ;
  window.push_back(_dlmin);
  int _dlmax=0;
  if ( ! (ifs >> _dlmax ) )   read_failure=true ;
  window.push_back(_dlmax);
  int _dcmin=0;
  if ( ! (ifs >> _dcmin ) )   read_failure=true ;
  window.push_back(_dcmin);
  int _dcmax=0;
  if ( ! (ifs >> _dcmax ) )   read_failure=true ;
  window.push_back(_dcmax);

  uint window_size= (_dlmax-_dlmin+1)*(_dcmax-_dcmin+1) ;
  for(uint i=0; i< window_size ;i++){
    double w=0.0;
    if ( ! (ifs >> w ) )   read_failure=true ;
    wghts.push_back(w);
  }

  ifs.close();

  if(read_failure){
    cerr << " --- Error in Image_Local_Neighborhood::ReadFromFile : " << endl;
    cerr << " --- Problem with file : " << filename << endl;
    cerr << " --- Possible file format error. Exiting." << endl;
    exit(1);
  }

  this->SetWeights( window, wghts);  
  this->precompute_neighbors();
}

/**

  Displays the object as  a generic neighborhood.

  Dumps the full structure of the neighborhood as if it was a generic one. 

  For verification purposes.

*/
   


void Image_Local_Neighborhood::DumpAsGeneric(){
  bool no_image=false;
  if((Nlines*Ncols)==0) {
    Nlines=(dlmax+dcmax-dlmin-dcmin+2);
    Ncols=Nlines;
    no_image=true;
  }
  uint nbpts=Nlines*Ncols;
  cout << 1 << endl;
  for(uint i=0; i< nbpts ;i++){
    vector< Weighted_Neighbor * >  ineighs;
    this->GetNeighbors(i,ineighs);
    uint nv=ineighs.size();
    cout << i << " " << nv << " ";
    for(uint j=0;j<nv;j++)  cout << ineighs[j]->Index << " ";
    for(uint j=0;j<nv;j++)  cout << ineighs[j]->Weight << " ";
    cout << endl;
    ineighs.clear();
  }
  
  if(no_image){
    Nlines=0;
    Ncols=0;
  }
}

/**
   Write the neighborhood definition to a text file. 

   \param [in] filename : name of the output file. 

   The format is described in \ref ReadFromFile .


*/

void Image_Local_Neighborhood::WriteToFile(string filename){

  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "Image_Local_Neighborhood::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool write_error=false;
  if ( !(ofs << dlmin << " " << dlmax << endl ) ) write_error=true;
  if ( !(ofs << dcmin << " " << dcmax << endl ) ) write_error=true;
  uint line_size=(dlmax-dlmin+1);
  
  for(uint i=0; i< weights.size() ;i++){
    if ( ! ( ofs << weights[i] << " " ) ) write_error=true;
    if(((i+1)%line_size)==0)
      if( ! ( ofs << endl ) ) write_error=true;
  }
  
  ofs.close();

  if(write_error){

    cerr << " --- Error : Image_Local_Neighborhood::WriteToFile : " << endl;
    cerr << " --- Problem with file : " << filename << endl;
    cerr << " --- Error writing file. Exiting." << endl;
    exit(1);
  }

}

void Image_Local_Neighborhood::Info(){
  cout << "-------------------------------------------- " << endl;
  cout << "Neighborhood structure information : " << endl;
  cout << "Type : Image local neighborhood : " << endl;
  cout << "Image Lines : " << Nlines << endl;
  cout << "Image Columns : " << Ncols <<  endl;

  uint nsz=0;
  for(uint i=0; i< weights.size() ;i++) if(weights[i]!=0) nsz++;
  cout << "Local neighborhood size : " << nsz << endl;
  
}
