#include <Stat.hpp>
#include <LinAlg.hpp>
#include <Utils.hpp>
#include <Normal.hpp>

Normal::Normal(){
  Dim=1;
  Mu.resize(Dim);
  Sigma.resize(Dim*Dim);
  for(uint d1=0;d1<Dim;d1++){
    Mu[d1]=0.0;
    for(uint d2=0;d2<Dim;d2++){
      if(d1==d2) Sigma[d1*Dim + d2 ]=1.0;
      else Sigma[ d1*Dim + d2]=0.0;	
    }
  }
	fix=false;    
}

Normal::Normal(uint dim){
  Dim=dim;
  Mu.resize(Dim);
  Sigma.resize(Dim*Dim);
  for(uint d1=0;d1<Dim;d1++){
    Mu[d1]=0.0;
    for(uint d2=0;d2<Dim;d2++){
      if(d1==d2) Sigma[d1*Dim + d2 ]=1.0;
      else Sigma[ d1*Dim + d2]=0.0;	
    }
  }
	fix=false;      
}

Normal::Normal(uint dim, vector<double> const &  mu, vector<double> const &  sigma){
  Dim=dim;
  Mu.resize(Dim);
  Sigma.resize(Dim*Dim);
  for(uint idx=0;idx<(Dim); idx++) Mu[idx]=mu[idx];
  for(uint idx=0;idx<(Dim*Dim); idx++) Sigma[idx]=sigma[idx];
	fix=false;
}

Normal::~Normal(){
  DeleteAll();
}

void Normal::DeleteAll(){
  Mu.clear();
  Sigma.clear();
  Sigma_inv_sqrt.clear();
}

void  Normal::Simulate(vector<double> & X){

  X.clear();
  rnorm(Mu,Sigma,X);
}


void Normal::Get_Mu(vector<double> &mu){
  mu.clear();
  mu.resize(Dim);
  for (uint d=0;d<Dim;d++) mu[d]=Mu[d];
}

void Normal::Get_Sigma(vector<double> &sigma){
  sigma.clear();
  sigma.resize(Dim*Dim);
  for (uint d=0;d<Dim*Dim;d++) sigma[d]=Sigma[d];
}


/// Precompute cholesky decomposition for all class covariance matrices

void  Normal::precompute_cholesky(){
  
  //cerr << "Normal::precompute_cholesky()"<< endl;
  uint Dim2=Dim*Dim;
  
  Sigma_inv_sqrt.clear();  
  Sigma_inv_sqrt.resize(Dim2);

  vector<double> L;

  // Sigma = L * L^T
  Cholesky(Sigma, L);
  // L1 = L^(-1)
  Matrix_invert(L,Sigma_inv_sqrt);

  // det = Det(L);
  Sigma_sqrt_det=Determinant(L);

  L.clear();
  //cerr << "end Normal::precompute_cholesky()"<< endl;

}

double Normal::LogDensity(vector<double> const &  X){
  double  LOG_TWO_PI=Log(2*M_PI);

  if( Sigma_inv_sqrt.size() != (Dim*Dim) ) precompute_cholesky();
  if(X.size() != Dim) {
    cerr << "Warning. wrong dimension in Normal::Density(vector<double> X)." << endl;
  }
  double logf=0.0;
  for (uint d1=0; d1< Dim ; d1++){
    double sum=0.0;
    for (uint d2=0; d2< Dim ; d2++){
      //sum += Sigma_inv_sqrt[ d1*Dim + d2] * (X[d1] - Mu[d1] );
			sum += Sigma_inv_sqrt[ d1*Dim + d2] * (X[d2] - Mu[d2] );
    }
    logf = logf - ( sum*sum + LOG_TWO_PI ) ;
  }

  logf = logf - 2.0 * Log( Sigma_sqrt_det ) ;
  logf = logf * 0.5;

  //  cout << "Normal::Density() " << logf << endl;
  return(logf);
}

void Normal::Estimate_Mu(Spatial_Data *spatd, vector<double>  const &  freqs){
  if (fix) return;
  //cerr << "Normal::Estimate_Mu()" << endl;
  // Number of sites
  uint N=spatd->Get_N();

  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  
  for(uint d=0 ; d < Dim ; d++){
    Mu[d]=0.0;
    for(uint i=0 ; i < N ; i++){
      double x_id=spatd->Get(i,d);	
      Mu[d]+=freqs[i]*x_id;
    }
    Mu[d]/=sum_freqs;
  }  
}



void Normal::Estimate_Sigma(Spatial_Data *spatd, vector<double>  const &  freqs){
  if (fix) return;
	//cerr << "Normal::Estimate_Sigma()" << endl;
  // Number of sites
  uint N=spatd->Get_N();
  
  double sum_freqs=0.0;
  for(uint i=0 ; i < N ; i++){
    sum_freqs+=freqs[i];
  }
  
  for(uint d1=0 ; d1 < Dim ; d1++){
    for(uint d2=d1 ; d2 < Dim ; d2++){     
      double s=0.0;
      for(uint i=0 ; i < N ; i++){
				double x_id1=spatd->Get(i,d1);
				double x_id2=spatd->Get(i,d2);
	
				s+=freqs[i]*(x_id1-Mu[d1])*(x_id2-Mu[d2]);
      }
      s/=sum_freqs;
      
      Sigma[d1*Dim + d2]= s;
      Sigma[d2*Dim + d1]= s;
    }
  }
  precompute_cholesky();
}


void Normal::Estimate(Spatial_Data *spatd, vector<double>  const &  freqs){
  Estimate_Mu(spatd,freqs);
  Estimate_Sigma(spatd,freqs);
}


void Normal::Estimate(Spatial_Data *spatd){
  spatd->Stats(Mu,Sigma);
}


void Normal::Read_Params(istream & is){
  ReadVector<double>(is,Mu);
  Dim=Mu.size();
  ReadVector<double>(is,Sigma);
  if(Sigma.size()!=(Dim*Dim)){
    cerr << "Normal::Read_Params() " << endl;
    cerr << "Error reading a Normal distribution parametrs." << endl;
    exit(0);
  }

}

void  Normal::Write_Params(ostream & os){
  WriteVector<double>(os,Mu);
  WriteVector<double>(os,Sigma);
}


void Normal::Set_Mu(vector<double> const &  m){
  for(uint d=0;d<Dim;d++) Mu[d]=m[d];  
}

void Normal::Set_Sigma(vector<double> const &  sig){
  for(uint d=0;d<(Dim*Dim);d++) Sigma[d]=sig[d];  
}

/// Number of free parameters
uint Normal::Degrees_Of_Freedom(){
  return(Dim + (Dim*(Dim+1))/2);
}

void Normal::Info(){
  //cout << "in void Normal::Info()" << endl;
  cout << "Normal:" << endl;
  cout << "Dim="<< Dim << endl;
  cout << "Mu="; WriteVector<double> (cout,Mu);
	cout << "Sigma="; WriteVector<double> (cout,Sigma);
}
