#include <Stat.hpp>
#include <LinAlg.hpp>
#include <Utils.hpp>
#include <Potts.hpp>

// default constructor
Potts::Potts(){
  Beta=0.1;
  K=1;
  NHS=new Image_Local_Neighborhood();
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();
  
  double step_init=1.8/( N * K * lmax );
  uint nb_param=1;
	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);
  
  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}

Potts::Potts(uint k){
  Beta=0.1;
  K=k;
  NHS=new Image_Local_Neighborhood();
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();
  double step_init=1.8/( N * K * lmax );  
  uint nb_param=1;
	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);

  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}

// Constructor
Potts::Potts(Neighborhood_System *nhs, uint k){
  Beta=0.1;
  K=k;
  NHS=nhs;
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();
  double step_init=1.8/( N * K  * lmax );  
  uint nb_param=1;
	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);
  
  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}

// Constructor
Potts::Potts(Neighborhood_System *nhs, uint k, double bet){
  Beta=bet;
  K=k;
  NHS=nhs;
  uint N=NHS->Get_N();
  uint lmax=NHS->LMAX();
  
  double step_init=1.8/( N * K  * lmax  );  
  uint nb_param=1;
	gradient_step.clear();
	gradient_step.resize(nb_param,step_init);  
  
  gradient_maxiter=2000;
  gradient_tol=0.02;
  precomputed_smf=false;
	fix=false;
}

double Potts::Get_Beta(){
  return(Beta);
}

void Potts::Get_Beta(vector<double> & beta){
	beta.clear();
	beta.resize(1,Beta);
}


double Potts::H(uint i , uint k, vector<uint>  const &  Z){

  vector< Weighted_Neighbor * >  ineighs;
  ineighs.clear();
  NHS->GetNeighbors(i,ineighs);
  uint nv=ineighs.size();

  double pot_c=0.0;
  for (uint v = 0; v<nv ; v++){
    
    uint    j   = ineighs[v]->Index;
    //double w_ij = ineighs[v]->Weight;
    
    double z_jk;
    if(Z[j]==k) z_jk=1;
    else z_jk=0;
    
    pot_c+= z_jk;
  }

  //  cout << i << "\t" << k << "\t"<< pot_c << endl;

  ineighs.clear();
  //  return(Beta*pot_c/nv);  
  return(Beta*pot_c);  
}


// default destructor
Potts::~Potts(){
  DeleteAll();
}


void Potts::DeleteAll(){
  
}


void Potts::ReadFromFile(string filename){
  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "External_Potts::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool read_error=false;
  
  if(!(ifs >> K)) read_error=true;  

  vector<double> b;
  ReadVector<double>(ifs,b);
  if(b.size()==1) Beta=b[0];
  if(b.size()!=1) read_error=true;

  b.clear();

  if(read_error){
    cerr << "External_Potts::ReadFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Possible format error. Exiting." << endl;
    exit(1);    
  }


}



void Potts::WriteToFile(string filename){
  ofstream ofs(filename.c_str());
  if(!ofs.good()){
    cerr << "External_Potts::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool write_error=false;
  
  if(!(ofs << K)) write_error=true;
  ofs << endl;
  vector<double> b(1);
  b[0]= Beta;
  WriteVector<double>(ofs,b);

  b.clear();

  if(write_error){
    cerr << "External_Potts::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Possible format error. Exiting." << endl;
    exit(1);    
  }

}




double Potts::Hmf(uint i, uint k){
  double e = 0.0;
  if( precomputed_smf )  e = Beta * S_mf[i*K+k];
  else{
    vector< Weighted_Neighbor * >  ineighs;
    NHS->GetNeighbors(i,ineighs);
    uint nv=ineighs.size();
    double sum=0.0;
    for(uint j=0;j<nv;j++){
      uint neidx=ineighs[j]->Index;
      sum += Z_mf[neidx*K+k];
    }
    ineighs.clear();
    e=Beta * sum;
  }
  return(e);
}


void Potts::Gradient(vector<double> const &  t, vector<double> & Q){

  //cerr << "Potts::Gradient() ." << endl;

  if(!precomputed_smf) Precompute_Smf();

  uint N=NHS->Get_N();

  Q.clear();
  Q.resize(1);
  
  
  
  vector<double> P_G;
  Compute_PGmf(P_G);
  double sum_i=0;
  for(uint p=0; p<K ; p++){
    for(uint i=0;i<N;i++){
      double sum_zip=S_mf[i*K+p];
      double gip= sum_zip*(t[i*K+p] - P_G[i*K+p]);
      sum_i += gip;
    }
  }
  Q[0]=sum_i;       

  P_G.clear();
}

void Potts::Update_Gradient(vector<double>  const &  Q){

  Beta = Beta + gradient_step[0] * Q[0];
  //  cout << "Potts::Update_Gradient() . Beta = " << Beta << endl; 
}


/// Number of free parameters
uint Potts::Degrees_Of_Freedom(){
  return(1);
}



void  Potts::Read_Params(istream & is){
  vector<double> b;
  ReadVector<double>(is,b);
  if(b.size()==1) Beta=b[0];
  
  if( (K<=0) || (b.size()!=1) ){
    cerr << "Potts::Read_Params() " << endl;
    cerr << "Error reading  parametrs." << endl;
    exit(0);
  }
  b.clear();
}

void  Potts::Write_Params(ostream & os){
  vector<double> b(1);
  b[0]= Beta;
  WriteVector<double>(os,b);
  b.clear();
}

void Potts::Info(){
  cout << "Standard Potts :" << endl;
  cout << "K = " << K << endl;
  cout << "Beta = " << Beta << endl;
}
