#include <Stat.hpp>
#include <LinAlg.hpp>
#include <Utils.hpp>
#include <Random_Field.hpp>

/**

Destructor.

*/
Random_Field::~Random_Field(){
  Z_mf.clear();
	S_mf.clear();
	gradient_step.clear();
}

void Random_Field::Fix(){
	fix=true;
}

/**

Precompute S_mf_i= \sum_{j \in N_i} Zmf_{jk}, \forall i, \forall k
This quantity does not change during the EM steps of the NR-EM algorithm.

*/


void Random_Field::Precompute_Smf(){
	cerr << "Warning! Random_Field::Precompute_Smf() is virtual." << endl;
}
  

/** Get the total number of classes

@return the total number of classes in the Random_Field.

*/

uint Random_Field::Get_K(){
	cerr << "Warning! uint Random_Field::Get_K() is virtual." << endl;
	return 0;
}



/**
Get the number of sites.

@return The number of sites.

*/


uint Random_Field::Get_N(){
  return(NHS->Get_N());
}

/**

Compute the Hamiltonian 

@param  i : a site.
@param  k : a class.
@param  z : current state.

@return The value of the hamiltonian.

Abstract. Model-dependent.

*/


double Random_Field::H(uint i, uint k, vector<uint> const & z){
  cerr << "Warning! Random_Field::H(uint i, uint k, vector<uint> const & z) is virtual." << endl;
  return(0.0);
}

/**

Compute the Hamiltonian mean-field approximation, given \f$ \tilde Z \f$

@param  i : a site.
@param  k : a class.

@return The value of the approximate hamiltonian.


Abstract. Model-dependent.

*/



double Random_Field::Hmf(uint i, uint k){
  cerr << "Warning! Random_Field::Hmf(uint i, uint k) is virtual." << endl;
  return(0.0);
}


/**
   Compute 
   
   \f$ P(Z_i = k | \tilde z, Y_i=p  ) = \frac {Exp[ Hmf_ipk]} { \sum_h Exp [ Hmf_iph] } \f$

  @param[out] PG : a vector containing all the probabilities.

*/

void Random_Field::Compute_PGmf(uint i, vector<double> & PG){
  uint K=Get_K();

  PG.clear();
  PG.resize(K,0);

	double max=0.0;
 	for(uint k=0;k<K;k++){
    double H_ik=Hmf(i,k);
  	if (H_ik>max) max=H_ik;
	}
	
  double sum=0.0;
  for(uint k=0;k<K;k++){
    double H_ik=Hmf(i,k)-max;
    PG[k]=Exp(H_ik);
    sum+=PG[k];
    //      cout << "H_ik= " << H_ik <<  "\t Exp(H_ik)= " <<  PG[i*K+k] << endl;
  }
  for(uint k=0;k<K;k++)  PG[k] /= sum;
}

void Random_Field::Compute_PGmf(vector<double> & PG){
	uint K=Get_K();
	uint N=Get_N();
	PG.clear();
	PG.resize(N*K);
	for (uint i=0;i<N;i++){
		vector<double> PG_i;
		Compute_PGmf(i,PG_i);
		for (uint k=0;k<K;k++) PG[i*K+k]=PG_i[k];
	}
}

/**

Compute the conditional probabilities:

\f$ P[Z_i=k | Z_{N_i}, Y_i=l ] = \frac {H_ilk} {\sum_h H_ilh} $


 @param[in] Z : current state.
 @param  i : a site.

  For gibbs sampler, Pseudo-Likelihood, ...
*/


void Random_Field::Compute_PG(uint i, vector<uint>  const &  Z, vector<double> &  PG){
  uint K=Get_K();

  PG.clear();
  PG.resize(K,0);

	double max=0.0;
  for (uint k=0; k<K ; k++) {
    double H_ik=H(i,k,Z);   
		if (H_ik>max) max=H_ik;
	}
	
  double sum_i=0.0;
  for (uint k=0; k<K ; k++) {
    double H_ik=H(i,k,Z)-max;   
    double num_k = Exp(H_ik);
    //    cerr << "H_ik = " << H_ik <<  " num_k = " << num_k << endl;
    sum_i += num_k; 
    PG[k]=num_k;
  }
  //  cerr << "sum_i = " << sum_i << endl;
  
  for (uint k = 0; k < K ; k++) PG[k]/=sum_i;

}


/**

 Simulation algorithm. Gibbs Sampler. 1 iteration over all the sites.

 @param[in] Z : current state.
 @param[out] Z : new state

*/

void Random_Field::Simulate(vector<uint> & Z){
  Simulate(1,Z);
}

    


/**

 Simulation algorithm. Gibbs Sampler. \b N iterations over all the sites.

 @param[in] nbiter : number of iterations.
 @param[in] Z : current state.
 @param[out] Z : new state


*/


void Random_Field::Simulate(uint nbiter, vector<uint> & Z){
  uint K=Get_K();
  uint N= NHS->Get_N();
  
  // Initialize uniformly if necessary (size condition)
  if(Z.size()!=N){
    Z.clear();
    Z.resize(N);
    for(uint i=0; i< N ; i++ ) Z[i] = rand() % K;
    cout << "Random_Field::Simulate(). Z.size()!=N " << endl;
  }

  for(uint it=0; it<nbiter;it++){
    
    for(uint i=0; i< N ; i++ ){
			vector<double> probs;
      Compute_PG(i,Z,probs);
      uint k=Sample(probs);
      Z[i]=k;
      probs.clear();
    }  

    if((it>=1000) && (it%1000) == 0 ) cerr << "Random_Field::Simulate(), iteration : " << it << endl;
  }

}


void Random_Field::Get_Zmf(vector<double> & z){

  uint K=Get_K();
  uint N=NHS->Get_N();
  if(z.size()!=(N*K)){
    z.clear();
    z.resize(K*N);
  }
  for(uint i=0; i<N; i++){
    for(uint k=0; k<K; k++){
      z[i*K+k]=Z_mf[i*K+k];
    }
  }
  
}

/**

 Set the mean-field neighborhood, \f$ \tilde Z \f$

 @param[in] z : vector of doubles.


 

*/

void Random_Field::Set_Zmf( vector<double> const &  z ){
  uint N=NHS->Get_N();
  uint K=Get_K();

  if(Z_mf.size()!=(N*K)){
    Z_mf.clear();
    Z_mf.resize(K*N);
  }
  for(uint idx=0; idx< (K*N); idx++) Z_mf[idx]=z[idx];
}


/**

 Set the mean-field neighborhood, \f$ \tilde Z \f$ , at site i

 @param[in] z : vector of doubles.
 @param[in] i : a site.


 

*/
void Random_Field::Set_Zmf(vector<double> const &  z, uint i){
  uint K=Get_K();
  //  cerr << "Normal_Random_Field::Set_Zmf() ." << endl;
  for(uint k=0; k< K; k++) Z_mf[i*K+k]=z[k];
}


/**

 Set the mean-field neighborhood, given Random_Field current state \f$ Z \f$.

 @param[in] z : current Random_Field state.

 Define|Set neighborhod conditions.
 

*/


void Random_Field::Set_Z( vector<uint> const &  z ){
  uint N=NHS->Get_N();
  uint K=Get_K();
  //  cout << N << endl;
  vector<double> zik(K*N);
  for(uint i=0; i<N; i++){
    for(uint k=0; k<K; k++){
      zik[i*K+k]=0.0;
    }
    zik[i*K+z[i]]=1.0;
  }
  //  Summary(zik,K);
  Set_Zmf(zik);
  zik.clear();
}


void Random_Field::Init_Gradient_Descent(){
  Precompute_Smf();  
  precomputed_smf=true;
}

void Random_Field::Finalize_Gradient_Descent(){
  S_mf.clear();
  precomputed_smf=false;
}

void Random_Field::Update_Gradient_Step(vector<double> const & Qold, vector<double> const & Q){
	//cout << "Random_Field::Update_Gradient_Step()" << endl;
  if (Qold.size()!=Q.size()){
    cerr << "Warning  Random_Field::Update_Gradient_Step: Qold.size()!=Q.size()"<<endl;
    cerr << "Exiting."<< endl;
  }
  else{
    for (uint s=0;s<Q.size();s++){
      if (Qold[s]*Q[s]>0) gradient_step[s]*=1.005;//1.5;
      else gradient_step[s]/=1.01;//2;
	//LA ON A lES PARAMETRES QUI DISENT COMMENT ON PEUT ACCELERER L'ALGO...
    }
  }
}


/**
 Performs a gradient descent to estimate model parameters.

 @param[in] tik : fuzzy classification.

*/


void Random_Field::Gradient_Descent(vector<double> const &  tik){
  vector<double> Q;

  Init_Gradient_Descent();
  Gradient(tik, Q);
  double norm_q=norm(Q);
  uint iter=0;
  vector<double> Qold(Q.size(),0);
  while( (norm_q> gradient_tol) && ( iter < gradient_maxiter ) ){
		//cout <<"***iter=" << iter << endl;
    Update_Gradient_Step(Q,Qold);
    Update_Gradient(Q);
    Qold.clear();
    Qold.resize(Q.size());
    for (uint s=0;s<Qold.size();s++) Qold[s]=Q[s];
    Gradient(tik, Q);
    norm_q=norm(Q);
    iter++;
    if (iter%100==0) {
			cout << "Random_Field::Gradient_Descent : Iteration :" << iter << " norm(Q) = " << norm_q << endl; 
			//WriteVector<double>(cout,Q);
		}
  }
  cerr << "Random_Field::Gradient_Descent : " << iter  << " iterations. norm(Q) = " <<  norm_q << endl;
  Finalize_Gradient_Descent();
}


void Random_Field::Set_Gradient(double step, double tol, uint maxiter){
  gradient_tol=tol;
  gradient_maxiter=maxiter;
  for (uint s=0;s<gradient_step.size();s++) gradient_step[s]=step;
}

void Random_Field::Update_Gradient(vector<double> const &   Q) {
	cerr << "Warning! Update_Gradient(vector<double> const &   Q) is virtual." << endl;
}

/**

Computes the gradient given the fuzzy classification.


 @param[in] t : fuzzy classification.
 @param[out] Q : Gradient vector.


*/
void Random_Field::Gradient(vector<double> const & tik, vector<double> & Q) {
	cerr << "Warning! Gradient(vector<double> const & tik, vector<double> & Q) is virtual." << endl;
}

/**

Computes the gradient given current Random_Field state.


 @param[in] Z : Random_Field state.
 @param[out] Q : Gradient vector.


*/

void Random_Field::Gradient(vector<uint> const &  Z, vector<double> & Q){
  
  Set_Z(Z);
  Gradient(Z_mf,Q);
}


/**
 Estimate the parameters given the state \f$ Z \f$
 
 @param[in] Z : current Random_Field state.
 
 
*/

void Random_Field::Estimate(vector<uint> const &  Z){
  uint N=NHS->Get_N();
	uint K=Get_K();
  vector<double> tik(K*N);
  for(uint i=0; i<N; i++){
    for(uint k=0; k<K; k++){
      tik[i*K+k]=0.0;
    }
    tik[i*K+Z[i]]=1.0;
  }
  Set_Zmf(tik);
  Estimate(tik);
  tik.clear();
}

/**
 Estimate the parameters given the fuzzy classification \f$ t_ik \f$.
 
 @param[in] tik : fuzzy classification.
 
 Performs a gradient descent to estimate model parameters.

*/



void Random_Field::Estimate(vector<double> const &  tik){
  if (!fix) Gradient_Descent(tik);
}


uint Random_Field::Degrees_Of_Freedom(){

  cerr << "Warning! Random_Field::Degrees_Of_Freedom() is virtual." << endl;
  
  return(0);

}


/**
Read the model parameters from the stream \b is.

*/

void Random_Field::Read_Params(istream & is){
  cout << "Random_Field::Read_Params() :  not defined."  << endl;
}

/**

Write the model parameters to the stream os.

*/

void Random_Field::Write_Params(ostream & os){
  cout << "Random_Field::Write_Params():  not defined."  << endl;

}

/**
   
Print some info on the model.

*/

void Random_Field::Info(){
  cout << "Random_Field::Info() :  not defined."  << endl;
}

