#include <Spatial_Data.hpp>
#include <Utils.hpp>


/// Neighborhood type marker in the structure file.
char NeighborhoodTypeMark[ TYPE_NB ] = { 'S', 'I', 'N' } ;


/**
   Default constructor.

   Creates a spatial dataset with a default dataset and a default neighborhood system.

*/

Spatial_Data::Spatial_Data(){
  data=new Data();
  nhsys=new Neighborhood_System();
}

/**
   Default constructor.

   \param [in] d : \ref Data object
   \param [in] nhs : \ref Neighborhood_System object.

   Creates a spatial dataset with a given  dataset \b dat and a given neighborhood system \b nhs.

*/

Spatial_Data::Spatial_Data(Data *d, Neighborhood_System *nhs){
  data=d;
  nhsys=nhs;
}

/**
   Deletes the object.
   
   Calls \ref DeleteAll

*/


Spatial_Data::~Spatial_Data(){
  DeleteAll();
}


/**
  Frees all the memory used by the object.
*/

void Spatial_Data::DeleteAll(){
  delete(data);
  delete(nhsys);
}

/**

   Get the neighborhood system.

*/


Neighborhood_System * Spatial_Data::Get_NS(){
  return(nhsys);
}


/**
Get the observation at the site i.

@param  i : a site.
@param  obs : a vector containing the observation at the site i.

*/

void Spatial_Data::Get(uint i, vector<double> & obs ){
  data->Get(i,obs);
}


/**

 Get X_i[d]

*/

double Spatial_Data::Get(uint i, uint d){
  double x=data->Get(i,d);
  return(x);
}



/**
Get the label at the site i.

@param  i : a site.
@return the label at the site i.

*/

uint Spatial_Data::GetLabel(uint i){
  
  uint lab=data->GetLabel(i);
  return(lab);
}


/**
Get the labels.

@return the labels at all the sites.

*/


void Spatial_Data::GetLabels(vector<uint> & labs){
  data->GetLabels(labs);
}




/**
Get the number of observations, i.e. , the size of the dataset.

@return The number of observations in the dataset.

*/

uint Spatial_Data::Get_N(){
  return(data->Get_N());
  
}



/**
Get the dimension.

@return The dimension of the observations.

*/

uint Spatial_Data::Get_D(){
  return(data->Get_D());
}


/** 
  Get the (weighted) neighbors for a given site (in the underlying neighborhood system). 

  @param[in] site : a site.
  @param[out]  neighs : the set of neighbors of the site \b site

  Fills the vector \b neighs with the necessary neighborhood data for the site \b site.


*/


void Spatial_Data::GetNeighbors(uint site, vector< Weighted_Neighbor * > & neighs){
  nhsys->GetNeighbors(site, neighs);
}



/** 
    Read the structure of 5the spatial dataset in a file.
    
    \param [in] filename : name of the input file. 

    File format is as follows :

    On a unique line :

    \b type \b size \b dim

    where :
    
    <ul>
      <li> \b type is the neighborhood type marker. one of   \b N , \b I , \b S :
      <ul>
      <li> \b N = Non-spatial neighborhood (TBD)
      <li>   \b S = ( Spatial ) generic neighborhood.
      <li>   \b I = ( Image ) local neighborhood.
      </ul>
     <li> \b size is :
     <ul>
           <li> \b NbPts  : if (\b type == \b N or \b S)
	    <li> \b Nlines \b Ncols  :  if ( \b type == \b I )
        </ul>  
     <li>  \b dim : is the dimension (number of variables).
   </ul>
   A typical structure file :

   \code
   S 65536 4
   \endcode

   or 

   \code
   I 256 256 4
  \endcode

*/

void Spatial_Data::ReadStructureFromFile(string filename){
  // read ".str" header file
  // test if file exists

  DeleteAll();

  ifstream ifs(filename.c_str());
  if(!ifs.good()){
    cerr << "Spatial_Data::ReadStructureFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }

  bool read_error=false;
  bool format_error=false;

  // First, read opening comments;
  string textline;

  while(!ifs.eof()){
    if( ! getline (ifs, textline) ) read_error=true;
    if ( (textline.length()>0) && (textline[0]=='#')) ; // skip comments
    else if(textline.length() == 0) ; // skip empty lines.
    else break;    
    //   cerr << "Comment  : " << textline << endl;
  }
  //  cerr << "Read : " << textline << endl;

  if(textline.length() == 0) format_error=true;
  
  if( ! ( read_error || format_error ) ){
    char Neighborhood_type=textline[0];
    
    string spec_text=textline.substr(1,textline.length()-1);
    //  cerr << "spec_text : " << spec_text << endl;
    std::istringstream iss(spec_text);
    
    //  cerr << "Neighborhood_type = " << Neighborhood_type << endl;
    Neighborhood_type=upper( Neighborhood_type );
    //  cerr << "Neighborhood_type = " << Neighborhood_type << endl;
    
    if(Neighborhood_type == NeighborhoodTypeMark[TYPE_NONSPATIAL]){
      /* Spatial irregular */
      uint nbpts=0;
      uint nbvars=0;
      if ( ! ( iss >> nbpts ) ) read_error=true;
      if ( ! ( iss >> nbvars ) ) read_error=true;    
      nhsys=new Generic_Neighborhood_System();
      data=new Data(nbpts,nbvars);
    }
    else if(Neighborhood_type == NeighborhoodTypeMark[TYPE_SPATIAL]){
      /* Generic neighborhood system */
      uint nbpts=0;
      uint nbvars=0;
      if ( ! ( iss >> nbpts ) ) read_error=true;
      if ( ! ( iss >> nbvars ) ) read_error=true;    
      nhsys=new Generic_Neighborhood_System();
      data=new Data(nbpts,nbvars);
    }
    else if(Neighborhood_type == NeighborhoodTypeMark[TYPE_IMAGE]){
      /* image spatial data */
      uint Nl=0;
      uint Nc=0;
      uint nbvars=0;
      if ( ! ( iss >> Nl ) ) read_error=true;
      if ( ! ( iss >> Nc ) ) read_error=true;
      if ( ! ( iss >> nbvars ) ) read_error=true;    
      nhsys=new Image_Local_Neighborhood(Nl,Nc);
      data=new Data(Nl*Nc,nbvars);    
    }
    else {
      format_error=true;
    }
  }
  
  if( read_error || format_error ){
    cerr << "Spatial_Data::ReadStructureFromFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Possible format error. Exiting." << endl;
    exit(1);
  }  

}

/**
   Read the embedded dataset from a file.

    \param [in] filename : name of the input file. 

    
    Calls Data::ReadFromFile

*/

void Spatial_Data::ReadDataFromFile(string filename){
  // read data
  data->ReadFromFile(filename);
}

/**
   Read the embedded neighborhood system from a file.

    \param [in] filename : name of the input file. 

    
    Calls Neighborhood_System::ReadFromFile

*/



void Spatial_Data::ReadNeighborhoodFromFile(string filename){
  // read nhsys
  nhsys->ReadFromFile(filename);
}

/**
   Read data and neighborhood system from a file.

    \param [in] filename : name of the input file (without extension). 

*/



void Spatial_Data::ReadFromFile(string filename){
  
  this->ReadStructureFromFile(filename+".str");
  nhsys->ReadFromFile(filename+".nei");
  data->ReadFromFile(filename+".dat");
  
}


/**
   Write data and neighborhood system to a file.

    \param [in] filename : name of the output file (without extension). 

*/



void Spatial_Data::WriteToFile(string filename){
  
  //  S , N , D
  string strfname=filename+".str";
  ofstream ofs(strfname.c_str());
  if(!ofs.good()){
    cerr << "Spatial_Data::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else    {
    ofs << "S " << Get_N() << " " << Get_D() << endl;
  }
  ofs.close();

  data->WriteToFile(filename+".dat");

  //  nhsys->WriteToFile(filename+".nei");

  string neifname=filename+".nei";
  ofstream ofs1(neifname.c_str());
  if(!ofs1.good()){
    cerr << "Spatial_Data::WriteToFile : " << endl;
    cerr << "Problem with file : " << filename << endl;
    cerr << "Exiting." << endl;
    exit(1);
  }
  else {
    ofs1 << 1 << endl;
    uint N=Get_N();
    for(uint i=0; i< N ;i++){
      vector< Weighted_Neighbor * >  ineighs;
      this->GetNeighbors(i,ineighs);
      uint nv=ineighs.size();
      ofs1 << i << " " << nv << " ";
      for(uint j=0;j<nv;j++)  ofs1 << ineighs[j]->Index << " ";
      for(uint j=0;j<nv;j++)  ofs1 << ineighs[j]->Weight << " ";
      ofs1 << endl;
      ineighs.clear();
    }
  }
  ofs1.close();
}
  





/**
   Compute simple statistics from the data (mean and covariance).
   
   Useful for estimating a gaussain distribution.

*/

void  Spatial_Data::Stats(vector<double> & mean, vector<double> & cov){

  data->Stats(mean,cov);

}

/**
 K-means algorithm
*/
void Spatial_Data::K_means(uint K, vector<double> & centers, vector<uint> & labels){

  data->K_means( K, centers,labels);

}



/**
   Display some information about the spatial dataset.

   Calls Data::Info()

   Calls Neighborhood_System::Info()

*/

void Spatial_Data::Info(){
  data->Info();
  nhsys->Info();
}
