#include<Stat.hpp>
#include<Utils.hpp>
#include<LinAlg.hpp>

/**
   exponential function.

   @param[in] x : double
   
   Avoids numerical errors for very small negative values of x as well as very large positive values of x.

*/


double Exp(double x){
  double res=0.0;
  double maxexp=600.0;
  if(x<=-maxexp){
    //    cerr << "Warning. exponential term is too low. " << x << " < -" << maxexp << endl; 
    res=1.0/exp(maxexp);
    
  }
  if(x>=maxexp){
    //cerr << "Warning. exponential term is too high.  " << x << " > " << maxexp << endl; 
    res=exp(maxexp);
  }
  if( (x>-maxexp) && (x<maxexp)) res=exp(x);
  return(res);
}

/**
   Logarithm function.

   @param[in] x : double

   Avoids numerical errors for small values of x.

*/


double Log(double x){
  double res=0.0;
  double maxexp=500.0;
  double minlog=1.0/exp(maxexp);
  if(x<=minlog){
    //    cerr << "Warning. Logarithmic term is too low.  " << x << " < " << minlog << endl;
    res=log(minlog);
  }
  else res=log(x);
  return(res);
}


/**
   compare indexes given the values in a vector. Used by weighted mean.
   

*/

class IndexedLessThan {
private:
  
  /// Vector used to sort indexes.
  vector<double> const  & table;
public:
  /// constructor
  IndexedLessThan( vector<double>  const &  aTable ) : table( aTable ) {}
  /// sort function
  bool operator() ( size_t a, size_t b ) const { return table[a] < table[b]; }

}; 

/**
   generate a random number uniformly in [0,1].

*/

double runif(){

  double uniform=drand48();
  return(uniform);
}


/**
   generate a random number following a gaussian distribution with mean 0 and variance 1 (uses Box and Muller method).

*/

double rnorm(){

  double U1=runif();
  double U2=runif();
  
  double theta=2*M_PI*U1;
  double R=sqrt(-2.0*Log(U2));

  double x=R*cos(theta);
  //  double y=R*sin(theta);

  return(x);
}


/**
   generate a random vector of n IID gaussian variables with mean 0 and variance 1 (uses Box and Muller method).

*/

void rnorm(uint n, vector<double> & result){

  result.clear();
  result.resize(n);
  for(uint i=0; i<(n/2) ; i++){
    double U1=runif();
    double U2=runif();
    
    double theta=2*M_PI*U1;
    double R=sqrt(-2.0*Log(U2));
    
    double x=R*cos(theta);
    double y=R*sin(theta);
    result[2*i]=x;
    result[2*i+1]=y;
  }

  if((n%2)!=0){
    result[n-1]=rnorm();
  }

}

/**
   generate a random vector of a d-dimensional normal distribution with mean mu_d and a diagonal covariance matrix sigma_d.

*/

void rnorm_diag(vector<double> const & mu, vector<double> const &  sigma,  vector<double> & result){
  uint D=mu.size();
  uint D2=sigma.size();
  if(D2!=D){
    cerr << "Error, normal distribtuion misspecification." << endl;
  }
  else{
    result.clear();
    result.resize(D);

    // normal iid vector
    vector<double> N;
    rnorm(D,N);

    for(uint d=0; d<D ;d++) result[d]=mu[d] + sqrt(sigma[d])*N[d];    

  }
}


/**
   generate a random vector of a d-dimensional normal distribution with mean mu_d and covariance matrix sigma_d.

*/

void rnorm(vector<double> const &  mu, vector<double> const &  sigma,  vector<double> & result){
  
  uint D=mu.size();
  uint D2=sigma.size();
  if(D2!=(D*D)){
    cerr << "Error, normal distribtuion misspecification." << endl;
  }
  else{
    result.clear();
    result.resize(D);
    
    // normal iid vector
    vector<double> N;
    rnorm(D,N);
    
    vector<double> L;
    Cholesky(sigma,L);
  
    vector<double> LN;
    Matrix_Vec_mult(L,N,LN);
    
    for(uint d=0; d<D ;d++) result[d]=LN[d]+mu[d];

    N.clear();
    L.clear();
    LN.clear();

  }
}




/** 
    Sampling from a distribution (p1, ...., pn)

    @param  [in] probs : vector containing the probabilities. 

    The vector paramter \b probs must be normalized.
    
*/

uint Sample(vector<double> const &  probs){
  uint n=probs.size();

  double uniform=runif();
  double CumSum=probs[0];

  uint k=0;
  
  while( (k<n) && (uniform > CumSum ) ) {
    k++;
    CumSum+=probs[k];
  }

  return(k);
}


/**
   Weighted median. For Laplace distribution mean estimation.

*/


double weighted_median(vector<double> const &  X , vector<double> const &  W){
 
  uint N=X.size();
  vector<uint> sort_index(N);
  for(uint i=0; i<N ; i++) sort_index[i]=i;
  sort(sort_index.begin(),sort_index.end(),IndexedLessThan( X ) );

  double sum_wi=0.0;
  for(uint i=0; i<N ; i++) sum_wi+=W[i];
  
  double weighted_med=0.0;

  double sum_j=0;
  for(uint j=0;j<N; j++){
    sum_j += W[ sort_index[j] ];
    if( sum_j == (0.5*sum_wi) ){
      weighted_med= 0.5 * ( X[ sort_index[j] ] + X[ sort_index[j+1] ] );
      break;
    }

    if( sum_j > (0.5*sum_wi) ){
      weighted_med= X[ sort_index[j] ] ;
      break;
    }
  }

  sort_index.clear();
  return(weighted_med);
  
}


/// choose k elements out of {0,1,....,N-1} without replacement.
void choose(uint N, uint K, vector<uint> & indexes){

  if(K<=N){
    indexes.clear();
    indexes.resize(K);
    
    vector<bool> already_selected(N);
    for(uint i=0 ; i < N ; i++) already_selected[i]=false;
    
    for(uint k=0; k<K ; k++){
      uint idx=rand() % (N-k);
      uint sel=0;
      //    uint out=0;
      uint j=0;
      while( (j<N) && (sel<=idx) ){
	if(!already_selected[j]) sel++;
	j++;
      }
      already_selected[j-1]=true;
      indexes[k]=j-1;
    }
  }
  else{
    cerr << "Choose(N,K) with K > N." << endl; 
  }

}


/// argmax
uint argmax(vector<double>  const & X){
  uint NX=X.size();
  uint k_max=0;
  for(uint k=1 ; k < NX ; k++){
    if(X[k_max]<=X[k]){
      k_max=k;
    }
  } 
  return(k_max);
}





/// Maximum
void Maximum(vector<double>  const & X, uint D, vector<double> & m){
  uint N=X.size()/D;
  m.clear();
  m.resize(D);
  for(uint d=0; d<D ; d++)  m[d]=X[d];
  
  for(uint d=0; d<D ; d++)
    for(uint i=0; i<N ; i++) {
      double val=X[i*D+d];
      if(val >= m[d]) m[d]=val;     
    }
}

/// Minimum
void Minimum(vector<double> const &  X, uint D, vector<double> & m){
  uint N=X.size()/D;
  m.clear();
  m.resize(D);
  for(uint d=0; d<D ; d++)  m[d]=X[d];
  
  for(uint d=0; d<D ; d++)
    for(uint i=0; i<N ; i++) {
      double val=X[i*D+d];
      if(val <= m[d]) m[d]=val;     
    }
}

/// Mean
void Mean_SD(vector<double> const &  X, uint D, vector<double> & m, vector<double> & s){
  uint N=X.size()/D;
  m.clear();
  m.resize(D);
  s.clear();
  s.resize(D);
  for(uint d=0; d<D ; d++)  m[d]=0.0;
  for(uint d=0; d<D ; d++)  s[d]=0.0;
  
  for(uint d=0; d<D ; d++)
    for(uint i=0; i<N ; i++) {
      double val=X[i*D+d];
      m[d]+=val;
      s[d]+=val*val;
    }

  for(uint d=0; d<D ; d++) m[d]/=N;
  for(uint d=0; d<D ; d++) s[d]/=N;
  for(uint d=0; d<D ; d++) s[d]-=m[d]*m[d];
  for(uint d=0; d<D ; d++) s[d]=sqrt(s[d]);

}

/// statistical summary
void Summary(vector<double> const &  X, uint D){
  uint N=X.size()/D;
  if(N>0){
    vector<double> m;
    vector<double> s;
    cout << "Maxima : " << endl;
    Maximum(X,D,m);
    WriteVector<double>(cout,m);
    cout << "Minima : " << endl;
    Minimum(X,D,m);
    WriteVector<double>(cout,m);
    Mean_SD(X,D,m,s);
    cout << "Means : " << endl;
    WriteVector<double>(cout,m);
    cout << "Standard deviations : " << endl;
    WriteVector<double>(cout,s);
    
  }
  else{
    cout << "Warning: Summary of empty vector." << endl;
  }
}
